# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from typing import Dict, Iterable, List, Optional, Tuple

from spinn_utilities.overrides import overrides

from pacman.model.graphs.application import (
    ApplicationFPGAVertex, FPGAConnection)
from pacman.model.graphs.common.slice import Slice
from pacman.model.graphs.machine import MachineFPGAVertex, MachineVertex
from pacman.model.routing_info import BaseKeyAndMask, RoutingInfo
from pacman.utilities.constants import BITS_IN_KEY
from pacman.utilities.utility_calls import get_n_bits

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.models.common import PopulationApplicationVertex
from .spif_devices import (
    SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK, SPIF_INPUT_FPGA_LINKS,
    N_PIPES, N_FIELDS, N_FILTERS, SpiNNFPGARegister, SPIFRegister,
    set_field_mask, set_field_shift, set_field_limit,
    set_filter_mask, set_filter_value, set_mapper_key,
    set_input_key, set_input_mask, set_input_route)


class SPIFInputDevice(
        ApplicationFPGAVertex, PopulationApplicationVertex,
        AbstractSendMeMulticastCommandsVertex):
    """ A 1D input device connected to SpiNNaker using a SPIF board.
    """

    #: SPIF outputs to 8 FPGA output links, so we split all input into 8
    INPUT_MASK = 7

    #: The number of devices in existence, to work out the key
    __n_devices = 0

    __slots__ = [
        "__spif_mask",
        "__index_by_slice",
        "__base_key",
        "__pipe",
        "__key_mask",
        "__m_vertex_mask",
        "__neuron_bits",
        "__index_bits",
        "__index_shift",
        "__input_mask",
        "__input_shift"]

    def __init__(self, pipe: int, n_neurons: int, n_neurons_per_partition: int,
                 base_key: Optional[int] = None,
                 board_address: Optional[str] = None,
                 chip_coords: Optional[Tuple[int, int]] = None):
        """

        :param pipe: Which pipe on SPIF the retina is connected to
        :param n_neurons: The number of neurons in the device
        :param n_neurons_per_partition: The number of neurons per partition
        :param base_key:
            The key that is common over the whole vertex,
            or None to use the pipe number as the key
        :param board_address:
            The IP address of the board to which the FPGA is connected, or None
            to use the default board or chip_coords.  Note chip_coords will be
            used first if both are specified, with board_address then being
            used if the coordinates don't connect to an FPGA.
        :param chip_coords:
            The coordinates of the chip to which the FPGA is connected, or
            None to use the default board or board_address.   Note chip_coords
            will be used first if board_address is also specified, with
            board_address then being used if the coordinates don't connect to
            an FPGA.
        """
        # Do some checks
        if n_neurons_per_partition < self.INPUT_MASK:
            raise ConfigurationException(
                "The number of neurons per partition must not be <"
                f" {self.INPUT_MASK} ({n_neurons_per_partition} specified)")
        if not self.__is_power_of_2(n_neurons_per_partition):
            raise ConfigurationException(
                f"n_neurons_per_partition ({n_neurons_per_partition}) "
                "must be a power of 2")
        if pipe >= N_PIPES:
            raise ConfigurationException(
                f"Pipe {pipe} is bigger than maximum allowed {N_PIPES}")

        n_machine_vertices = int(
            math.ceil(n_neurons / n_neurons_per_partition))

        # Call the super
        super().__init__(
            n_atoms=n_neurons,
            n_machine_vertices_per_link=n_machine_vertices,
            incoming_fpga_connections=self.__incoming_fpgas(
                board_address, chip_coords),
            outgoing_fpga_connection=self.__outgoing_fpga(
                board_address, chip_coords))

        # The mask is going to be made up of:
        # | K | N_I | N_0 | N_F |
        # K = base key
        # N_I = index of sub-partition
        # N_0 = 0s for neuron IDs not cared about
        # N_F = FPGA index
        # Now - go calculate:
        self.__neuron_bits = get_n_bits(n_neurons)
        self.__index_bits = get_n_bits(n_machine_vertices)
        self.__index_shift = get_n_bits(n_neurons_per_partition)

        # Mask to apply to route packets at input
        n_key_bits = BITS_IN_KEY - self.__neuron_bits
        self.__key_mask = ((1 << n_key_bits) - 1) << self.__neuron_bits
        self.__spif_mask = self.__key_mask + self.INPUT_MASK
        sub_mask = (1 << self.__index_bits) - 1
        self.__m_vertex_mask = (
            self.__key_mask + (sub_mask << self.__index_shift))

        # A dictionary to get vertex index from FPGA and slice
        self.__index_by_slice: Dict[Tuple[int, Slice], int] = dict()

        self.__pipe = pipe
        if base_key is None:
            self.__base_key = SPIFInputDevice.__n_devices
        else:
            self.__base_key = base_key
        self.__base_key = self.__base_key << self.__neuron_bits
        SPIFInputDevice.__n_devices += 1

        # Generate the shifts and masks to convert the SPIF Ethernet inputs to
        # output format
        self.__input_mask = (1 << self.__neuron_bits) - 1
        self.__input_shift = 0

    def __is_power_of_2(self, v: int) -> bool:
        """ Determine if a value is a power of 2

        :param v: The value to test
        """
        return (v & (v - 1) == 0) and (v != 0)

    def __incoming_fpgas(
            self, board_address: Optional[str],
            chip_coords: Optional[Tuple[int, int]]) -> List[FPGAConnection]:
        """ Get the incoming FPGA connections  """
        # We use every other odd link
        return [FPGAConnection(SPIF_FPGA_ID, i, board_address, chip_coords)
                for i in SPIF_INPUT_FPGA_LINKS]

    def __outgoing_fpga(
            self, board_address: Optional[str],
            chip_coords: Optional[Tuple[int, int]]) -> FPGAConnection:
        """ Get the outgoing FPGA connection (for commands) """
        return FPGAConnection(
            SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK, board_address, chip_coords)

    def __fpga_index(self, fpga_link_id: int) -> int:
        # We use every other odd link, so we can work out the "index" of the
        # link in the list as follows
        return (fpga_link_id - 1) // 2

    @overrides(ApplicationFPGAVertex.get_incoming_slice_for_link)
    def get_incoming_slice_for_link(
            self, link: FPGAConnection, index: int) -> Slice:
        vertex_slice = super().get_incoming_slice_for_link(link, index)
        self.__index_by_slice[link.fpga_link_id, vertex_slice] = index
        return vertex_slice

    @overrides(ApplicationFPGAVertex.get_machine_fixed_key_and_mask)
    def get_machine_fixed_key_and_mask(self, machine_vertex: MachineVertex,
                                       partition_id: str) -> BaseKeyAndMask:
        assert isinstance(machine_vertex, MachineFPGAVertex)
        fpga_link_id = machine_vertex.fpga_link_id
        vertex_slice = machine_vertex.vertex_slice
        index = self.__index_by_slice[fpga_link_id, vertex_slice]
        key = self.__base_key | (index << self.__index_shift)

        fpga = self.__fpga_index(fpga_link_id)

        # Build the key from the components
        fpga_key = key + fpga
        fpga_mask = self.__m_vertex_mask | self.__spif_mask
        return BaseKeyAndMask(fpga_key, fpga_mask)

    @overrides(ApplicationFPGAVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id: str) -> BaseKeyAndMask:
        return BaseKeyAndMask(self.__base_key, self.__key_mask)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # Make sure everything has stopped
        commands = [SpiNNFPGARegister.STOP.cmd()]

        # Clear the counters
        commands.append(SPIFRegister.OUT_PERIPH_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.CONFIG_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.DROPPED_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.IN_PERIPH_PKT_CNT.cmd(0))

        # Configure the creation of packets from fields to keys using the
        # "standard" input to SPIF (X | P | Y) and convert to (Y | X)
        commands.extend([
            set_field_mask(self.__pipe, 0, self.__input_mask),
            set_field_shift(self.__pipe, 0, self.__input_shift),
            set_field_limit(self.__pipe, 0, self.n_atoms)])

        # These are unused but set them to be sure
        for i in range(1, N_FIELDS):
            commands.extend([
                set_field_mask(self.__pipe, i, 0),
                set_field_shift(self.__pipe, i, 0),
                set_field_limit(self.__pipe, i, 0)])

        # Don't filter
        commands.extend([
            set_filter_mask(self.__pipe, i, 0) for i in range(N_FILTERS)
        ])
        commands.extend([
            set_filter_value(self.__pipe, i, 1) for i in range(N_FILTERS)
        ])

        # Configure the output routing key
        commands.append(set_mapper_key(self.__pipe, self.__base_key))

        # Configure the links to send packets to the 8 FPGAs using the
        # lower bits; the input key is against the reversed links list because
        # the input numbering on SPIF is upside-down to the FPGA numbering, and
        # we put out the fixed key and mask in the FPGA link order, so we need
        # to match that with what SPIF will output.
        commands.extend(
            set_input_key(self.__pipe, i, self.__spif_key(f))
            for i, f in enumerate(reversed(SPIF_INPUT_FPGA_LINKS)))
        commands.extend(
            set_input_mask(self.__pipe, i, self.__spif_mask)
            for i in range(len(SPIF_INPUT_FPGA_LINKS)))
        commands.extend(
            set_input_route(self.__pipe, i, i)
            for i in range(len(SPIF_INPUT_FPGA_LINKS)))

        # Send the start signal
        commands.append(SpiNNFPGARegister.START.cmd())

        return commands

    def __spif_key(self, fpga_link_id: int) -> int:
        fpga = self.__fpga_index(fpga_link_id)
        return self.__base_key + fpga

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        # Send the stop signal
        return [SpiNNFPGARegister.STOP.cmd()]

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []

    @overrides(PopulationApplicationVertex.get_atom_key_map)
    def get_atom_key_map(
            self, pre_vertex: MachineVertex, partition_id: str,
            routing_info: RoutingInfo) -> Iterable[Tuple[int, int]]:
        # Work out which machine vertex
        start = pre_vertex.vertex_slice.lo_atom
        key_and_mask = self.get_machine_fixed_key_and_mask(
            pre_vertex, partition_id)
        end = pre_vertex.vertex_slice.hi_atom + 1
        n_key = key_and_mask.key & self.INPUT_MASK
        start += n_key
        neuron_id = pre_vertex.vertex_slice.lo_atom + n_key
        for n in range(start, end, self.INPUT_MASK + 1):
            key = key_and_mask.key | n
            yield (neuron_id, key)
            neuron_id += self.INPUT_MASK + 1



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Sequence
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractStandardNeuronComponent)


class AbstractSynapseType(
        AbstractStandardNeuronComponent, metaclass=AbstractBase):
    """
    Represents the synapse types supported.
    """

    __slots__ = ()

    @abstractmethod
    def get_n_synapse_types(self) -> int:
        """
        Get the number of synapse types supported.

        :return: The number of synapse types supported
        """
        raise NotImplementedError

    @abstractmethod
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        """
        Get the ID of a synapse given the name.

        :return: The ID of the synapse
        """
        raise NotImplementedError

    @abstractmethod
    def get_synapse_targets(self) -> Sequence[str]:
        """
        Get the target names of the synapse type.

        :return: an array of strings (usually a list or tuple)
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union
import numpy
from numpy.typing import NDArray


def synfire_spike_checker(spikes: Union[NDArray, List[NDArray]],
                          n_neurons: int) -> None:
    """
    Checks the spikes and prints an error file if needed.

    :param spikes: The spike data to check.
    :param n_neurons: The number of neurons.
    :raises Exception: If there is a problem with the data
    """
    if isinstance(spikes, numpy.ndarray):
        sorted_spikes = spikes[spikes[:, 1].argsort()]
        print(len(sorted_spikes))
        num = 0
        for row in sorted_spikes:
            if num != round(row[0]):
                numpy.savetxt("spikes.csv", sorted_spikes, fmt=['%d', '%d'],
                              delimiter=',')
                raise ValueError(f"Unexpected spike at time {row[1]}")
            num += 1
            if num >= n_neurons:
                num = 0
    else:
        for single in spikes:
            synfire_spike_checker(single, n_neurons)


def synfire_multiple_lines_spike_checker(
        spikes: NDArray, n_neurons: int, lines: int,
        wrap_around: bool = True) -> None:
    """
    Checks that there are the expected number of spike lines

    :param spikes: The spikes
    :param n_neurons: The number of neurons.
    :param lines: Expected number of lines
    :param wrap_around:
        If True the lines will wrap around when reaching the last neuron.
    :raises Exception: If there is a problem with the data
    """
    sorted_spikes = spikes[spikes[:, 1].argsort()]
    nums = [0] * lines
    used = [False] * lines
    for row in sorted_spikes:
        node = round(row[0])
        found = False
        for i in range(lines):
            if nums[i] == node:
                found = True
                nums[i] += 1
                if nums[i] >= n_neurons and wrap_around:
                    nums[i] = 0
                used[i] = True
                break
        if not found:
            numpy.savetxt("sorted_spikes.csv", sorted_spikes, fmt=['%d', '%d'],
                          delimiter=',')
            raise ValueError(f"Unexpected spike at time {row[1]}")
    if False in used:
        numpy.savetxt("sorted_spikes.csv", sorted_spikes, fmt=['%d', '%d'],
                      delimiter=',')
        print(used)
        raise ValueError(f"Expected {lines} spike lines")


if __name__ == '__main__':
    _spikes = numpy.loadtxt("sorted_spikes.csv", delimiter=',')
    synfire_multiple_lines_spike_checker(_spikes, 200, 10, wrap_around=False)
    # synfire_spike_checker(_spikes, 20)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, Tuple
from spinn_utilities.overrides import overrides
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from spynnaker.pyNN.models.common.types import Spikes
from .spike_source_array_vertex import SpikeSourceArrayVertex


class SpikeSourceArray(AbstractPyNNModel):
    """
    Model that creates a Spike Source Array Vertex
    """
    default_population_parameters = {
        "splitter": None, "n_colour_bits": None, "neurons_per_core": None}

    def __init__(self, spike_times: Optional[Spikes] = None):
        """
        :param spike_times: Timesteps on which to spike
        """
        if spike_times is None:
            spike_times = []
        self.__spike_times = spike_times

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *,
            splitter: Optional[AbstractSplitterCommon] = None,
            neurons_per_core: Optional[Union[int, Tuple[int, ...]]] = None,
            n_colour_bits: Optional[int] = None) -> SpikeSourceArrayVertex:
        """
        :param splitter:
        :param n_colour_bits:
        """
        if neurons_per_core is None:
            neurons_per_core = \
                self.get_model_max_atoms_per_dimension_per_core()
        return SpikeSourceArrayVertex(
            n_neurons, self.__spike_times, label, neurons_per_core, self,
            splitter, n_colour_bits)

    @property
    def _spike_times(self) -> Spikes:
        return self.__spike_times



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import ctypes
import math
from typing import (
    Dict, Iterable, List, Sequence, Tuple, Type, TypeVar, TYPE_CHECKING)

import numpy
from numpy import uint32
from numpy.typing import NDArray

from pacman.model.routing_info import BaseKeyAndMask

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.exceptions import (
    SynapseRowTooBigException, SynapticConfigurationException)
from spynnaker.pyNN.utilities.constants import POP_TABLE_MAX_ROW_LENGTH
from spynnaker.pyNN.utilities.bit_field_utilities import BIT_IN_A_WORD

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    # pylint: disable=no-member, protected-access
    _T = TypeVar("_T", bound=ctypes._CData)


# Scale factor for an address; allows more addresses to be represented, but
# means addresses have to be aligned to these offsets
_ADDRESS_SCALE = 16

# A padding byte
_PADDING_BYTE = 0xDD

# Bits in a byte
_BITS_PER_BYTES = 8

# ctypes stores the number of bits in a bitfield in the top 16 bits
_CTYPES_N_BITS_SHIFT = 16


def _n_bits(field: ctypes._CField) -> int:
    """
    Get the number of bits in a field (ctypes doesn't do this).

    :param field: a ctype field from a structure
    :return: the number of bits
    """
    # ctypes stores the number of bits in a bitfield in the top 16 bits;
    # if it isn't a bitfield, this is 0
    n_bits = field.size >> _CTYPES_N_BITS_SHIFT
    if n_bits:
        return n_bits

    # If it isn't a bitfield, the number of bits is the field size (which is
    # then in bytes) multiplied by 8 (bits in a byte)
    return _BITS_PER_BYTES * field.size


def _make_array(ctype: Type[_T], n_items: int) -> ctypes.Array[_T]:
    """
    Make an array of ctype items; done separately as the syntax is a
    little odd!

    :param ctype: A ctype
    :param n_items: The number of items in the array
    :return: a ctype array
    """
    array_type = ctype * n_items  # type: ignore
    return array_type()  # type: ignore


class _MasterPopEntryCType(ctypes.LittleEndianStructure):
    """
    A Master Population Table Entry; matches the C struct.
    """
    _fields_ = [
        # The key to match against the incoming message
        ("key", ctypes.c_uint32),
        # The mask to select the relevant bits of key for matching
        ("mask", ctypes.c_uint32),
        # The index into address_list for this entry
        ("start", ctypes.c_uint32, 13),
        # The number of colour bits used (or 0 if no colour bits)
        ("n_colour_bits", ctypes.c_uint32, 3),
        # The number of entries in ::address_list for this entry
        ("count", ctypes.c_uint32, 16),
        # The mask to apply to the key once shifted get the core index
        ("core_mask", ctypes.c_uint32, 16),
        # The shift to apply to the key to get the core part (0-31)
        ("mask_shift", ctypes.c_uint32, 16),
        # The number of neurons per core
        ("n_neurons", ctypes.c_uint32, 16),
        # The number of words required for n_neurons
        ("n_words", ctypes.c_uint32, 16)
    ]


# Maximum start position in the address list
_MAX_ADDRESS_START = (1 << _n_bits(_MasterPopEntryCType.start)) - 1
# Maximum count of address list entries for a single pop table entry
_MAX_ADDRESS_COUNT = (1 << _n_bits(_MasterPopEntryCType.count)) - 1

# The maximum n_neurons value
_MAX_N_NEURONS = (1 << _n_bits(_MasterPopEntryCType.n_neurons)) - 1
# Maximum core mask (i.e. number of cores)
_MAX_CORE_MASK = (1 << _n_bits(_MasterPopEntryCType.core_mask)) - 1


class _AddressListEntryCType(ctypes.LittleEndianStructure):
    """
    An Address and Row Length structure; matches the C struct.
    """
    _fields_ = [
        # the length of the row
        ("row_length", ctypes.c_uint32, 8),
        # the address
        ("address", ctypes.c_uint32, 24)
    ]


# An invalid address in the address and row length list
_INVALID_ADDDRESS = (1 << _n_bits(_AddressListEntryCType.address)) - 1
# Address is 23 bits, but maximum value means invalid
_MAX_ADDRESS = (1 << _n_bits(_AddressListEntryCType.address)) - 2

# Sizes of structs
_MASTER_POP_ENTRY_SIZE_BYTES = ctypes.sizeof(_MasterPopEntryCType)
_ADDRESS_LIST_ENTRY_SIZE_BYTES = ctypes.sizeof(_AddressListEntryCType)

# Base size - 2 words for size of table and address list
_BASE_SIZE_BYTES = 8

# Number of times to multiply for delays
_DELAY_SCALE = 2

# A ctypes pointer to a uint32
_UINT32_PTR = ctypes.POINTER(ctypes.c_uint32)


def _to_numpy(array: ctypes.Array) -> NDArray[uint32]:
    """
    Convert a ctypes array to a numpy array of uint32.

    .. note::
        No data copying is done; it is pure type conversion.  Editing
        the returned array will result in changes to the original.

    :param array: The array to convert
    """
    # Nothing to do if the array is 0 sized
    if len(array) == 0:
        return numpy.zeros(0, dtype=uint32)

    uint32_array = ctypes.cast(array, _UINT32_PTR)
    n_words = (len(array) * ctypes.sizeof(array[0])) // BYTES_PER_WORD
    return numpy.ctypeslib.as_array(uint32_array, (n_words,))


class _MasterPopEntry(object):
    """
    Internal class that contains a master population table entry.
    """
    __slots__ = (
        "__addresses_and_row_lengths",
        # The mask to match this entry on
        "__mask",
        # The routing key to match this entry on
        "__routing_key",
        # The part of the key where the core id is held after shifting (below)
        "__core_mask",
        # Where in the key that the core id is held
        "__core_shift",
        # The number of neurons on every core except the last
        "__n_neurons",
        # The number of bits reserved for the colour
        "__n_colour_bits")

    def __init__(self, routing_key: int, mask: int, core_mask: int,
                 core_shift: int, n_neurons: int, n_colour_bits: int):
        """
        :param routing_key: The key to match for this entry
        :param mask: The mask to match for this entry
        :param core_mask:
            The part of the routing_key where the core id is held
        :param core_shift: Where in the routing_key the core_id is held
        :param n_neurons:
            The number of neurons on each core, except the last
        """
        self.__routing_key = routing_key
        self.__mask = mask
        self.__core_mask = core_mask
        self.__core_shift = core_shift
        self.__n_neurons = n_neurons
        self.__n_colour_bits = n_colour_bits
        self.__addresses_and_row_lengths: List[Tuple[int, int, bool]] = list()

    def append(self, address: int, row_length: int) -> int:
        """
        Add a synaptic matrix pointer to the entry.

        :param address: The address of the synaptic matrix
        :param row_length: The length of each row in the matrix
        :return: The index of the pointer within the entry
        """
        index = len(self.__addresses_and_row_lengths)
        if index > _MAX_ADDRESS_COUNT:
            raise SynapticConfigurationException(
                f"{index} connections for the same source key "
                f"(maximum {_MAX_ADDRESS_COUNT})")
        self.__addresses_and_row_lengths.append(
            (address, row_length, True))
        return index

    def append_invalid(self) -> int:
        """
        Add an invalid marker to the entry; used to ensure index alignment
        between multiple entries when necessary.

        :return: The index of the marker within the entry
        """
        index = len(self.__addresses_and_row_lengths)
        self.__addresses_and_row_lengths.append((0, 0, False))
        return index

    @property
    def routing_key(self) -> int:
        """
        The key combo of this entry.
        """
        return self.__routing_key

    @property
    def mask(self) -> int:
        """
        The mask of the key for this entry.
        """
        return self.__mask

    @property
    def core_mask(self) -> int:
        """
        The mask of the key once shifted to get the source core ID.
        """
        return self.__core_mask

    @property
    def core_shift(self) -> int:
        """
        The shift of the key to get the source core ID.
        """
        return self.__core_shift

    @property
    def n_neurons(self) -> int:
        """
        The number of neurons per source core.
        """
        return self.__n_neurons

    @property
    def addresses_and_row_lengths(self) -> Sequence[Tuple[int, int, bool]]:
        """
        The memory address that this master pop entry points at
        (in the synaptic matrix).
        """
        return self.__addresses_and_row_lengths

    def write_to_table(
            self, entry: _MasterPopEntryCType,
            address_list: ctypes.Array[_AddressListEntryCType],
            start: int) -> int:
        """
        Write entries to the master population table.

        :param entry: The entry to write to
        :param address_list:
            The address_list to write to
        :param start:
            The index of the entry of the address list to start at
        :return: The number of entries written to the address list
        """
        entry.key = self.__routing_key
        entry.mask = self.__mask
        entry.start = start
        count = len(self.__addresses_and_row_lengths)
        entry.count = count

        # Mark where the next entry starts and the number added; this might
        # change if there is extra info
        next_addr = start
        n_entries = count

        entry.n_colour_bits = self.__n_colour_bits
        entry.core_mask = self.__core_mask
        entry.n_words = int(math.ceil(self.__n_neurons / BIT_IN_A_WORD))
        entry.n_neurons = self.__n_neurons
        entry.mask_shift = self.__core_shift

        for j, (address, row_length, is_valid) in enumerate(
                self.__addresses_and_row_lengths):
            address_entry = address_list[next_addr + j]
            if not is_valid:
                address_entry.address = _INVALID_ADDDRESS
            else:
                address_entry.row_length = row_length
                address_entry.address = address
        return n_entries


class MasterPopTableAsBinarySearch(object):
    """
    Master population table, implemented as binary search master.
    """
    __slots__ = (
        "__entries",
        "__n_addresses")

    def __init__(self) -> None:
        self.__entries: Dict[int, _MasterPopEntry] = {}
        self.__n_addresses = 0

    @staticmethod
    def get_master_population_table_size(
            incoming_projections: Iterable[Projection]) -> int:
        """
        Get the size of the master population table in SDRAM.

        :param incoming_projections:
            The projections arriving at the vertex that are to be handled by
            this table
        :return: the size the master pop table will take in SDRAM (in bytes)
        """
        # Count the pre-machine-vertices
        n_entries = 0
        n_vertices = 0
        seen_edges = set()
        for proj in incoming_projections:
            n_entries += 1
            # pylint: disable=protected-access
            in_edge = proj._projection_edge

            # Each projection with a delay will have an additional entry
            if in_edge.n_delay_stages:
                n_entries += 1

            # If we haven't seen this edge before, add it in
            if in_edge not in seen_edges:
                seen_edges.add(in_edge)
                # Each pre-vertex has a master pop entry
                n_vertices += 1
                if in_edge.n_delay_stages:
                    n_vertices += 1

        return (
            _BASE_SIZE_BYTES +
            (n_vertices * _MASTER_POP_ENTRY_SIZE_BYTES) +
            (n_entries * _ADDRESS_LIST_ENTRY_SIZE_BYTES))

    @staticmethod
    def get_allowed_row_length(row_length: int) -> int:
        """
        Get the next allowed row length.

        :param row_length: the row length being considered
        :return: the row length available
        :raises SynapseRowTooBigException: If the row won't fit
        """
        if row_length > POP_TABLE_MAX_ROW_LENGTH:
            raise SynapseRowTooBigException(
                POP_TABLE_MAX_ROW_LENGTH,
                f"Only rows of up to {POP_TABLE_MAX_ROW_LENGTH} "
                "entries are allowed")
        return row_length

    @staticmethod
    def get_next_allowed_address(next_address: int) -> int:
        """
        Get the next allowed address.

        :param next_address: The next address that would be used
        :return: The next address that can be used following next_address
        :raises SynapticConfigurationException:
            if the address is out of range
        """
        addr_scaled = (next_address + (_ADDRESS_SCALE - 1)) // _ADDRESS_SCALE
        if addr_scaled > _MAX_ADDRESS:
            raise SynapticConfigurationException(
                f"Address {hex(addr_scaled * _ADDRESS_SCALE)} is "
                "out of range for this population table!")
        return addr_scaled * _ADDRESS_SCALE

    def initialise_table(self) -> None:
        """
        Initialise the master pop data structure.
        """
        self.__entries = dict()
        self.__n_addresses = 0

    def add_application_entry(
            self, block_start_addr: int, row_length: int,
            key_and_mask: BaseKeyAndMask, core_mask: int,
            core_shift: int, n_neurons: int, n_colour_bits: int) -> int:
        """
        Add an entry for an application-edge to the population table.

        :param block_start_addr: where the synaptic matrix block starts
        :param row_length: how long in words each row is
        :param key_and_mask:
            the key and mask for this master pop entry
        :param core_mask:
            Mask for the part of the key that identifies the core
        :param core_shift: The shift of the mask to get to the core_mask
        :param n_neurons:
            The number of neurons in each machine vertex (bar the last)
        :param n_colour_bits:
            The number of bits to use for colour
        :return: The index of the entry, to be used to retrieve it
        :raises SynapticConfigurationException:
            If a bad address is used.
        """
        # If there are too many neurons per core, fail
        if n_neurons > _MAX_N_NEURONS:
            raise SynapticConfigurationException(
                f"The parameter n_neurons of {n_neurons} is too big "
                f"(maximum {_MAX_N_NEURONS})")

        # If the core mask is too big, fail
        if core_mask > _MAX_CORE_MASK:
            raise SynapticConfigurationException(
                f"The core mask of {core_mask} is too big "
                f"(maximum {_MAX_CORE_MASK})")

        return self.__update_master_population_table(
            block_start_addr, row_length, key_and_mask, core_mask, core_shift,
            n_neurons, n_colour_bits)

    def __update_master_population_table(
            self, block_start_addr: int, row_length: int,
            key_and_mask: BaseKeyAndMask, core_mask: int, core_shift: int,
            n_neurons: int, n_colour_bits: int) -> int:
        """
        Add an entry in the binary search to deal with the synaptic matrix.

        :param block_start_addr: where the synaptic matrix block starts
        :param row_length: how long in words each row is
        :param key_and_mask: the key and mask for this master pop entry
        :param core_mask:
            Mask for the part of the key that identifies the core
        :param core_shift: The shift of the mask to get to the core_mask
        :param n_neurons:
            The number of neurons in each machine vertex (bar the last)
        :param n_colour_bits: The number of bits to use for colour
        :return: The index of the entry, to be used to retrieve it
        :raises SynapticConfigurationException:
            If a bad address is used.
        """
        # if not single, scale the address
        if block_start_addr % _ADDRESS_SCALE != 0:
            raise SynapticConfigurationException(
                f"Address {block_start_addr} is not compatible "
                "with this table")
        start_addr = block_start_addr // _ADDRESS_SCALE
        if start_addr > _MAX_ADDRESS:
            raise SynapticConfigurationException(
                f"Address {block_start_addr} is too big for this table")
        row_length = self.get_allowed_row_length(row_length)

        entry = self.__add_entry(
            key_and_mask, core_mask, core_shift, n_neurons, n_colour_bits)
        index = entry.append(start_addr, row_length - 1)
        self.__n_addresses += 1
        return index

    def add_invalid_application_entry(
            self, key_and_mask: BaseKeyAndMask, core_mask: int,
            core_shift: int, n_neurons: int, n_colour_bits: int) -> int:
        """
        Add an entry to the table from an application vertex that doesn't
        point to anywhere.  Used to keep indices in synchronisation between
        e.g. normal and delay entries and between entries on different cores.

        :param key_and_mask:
            a key_and_mask object used as part of describing
            an edge that will require being received to be stored in the
            master pop table; the whole edge will become multiple calls to
            this function
        :param core_mask:
            Mask for the part of the key that identifies the core
        :param core_shift: The shift of the mask to get to the core_mask
        :param n_neurons:
            The number of neurons in each machine vertex (bar the last)
        :param n_colour_bits:
            The number of bits to use for colour
        :return: The index of the added entry
        """
        # If there are too many neurons per core, fail
        if n_neurons > _MAX_N_NEURONS:
            raise SynapticConfigurationException(
                f"The parameter n_neurons of {n_neurons} is too big "
                f"(maximum {_MAX_N_NEURONS})")

        # If the core mask is too big, fail
        if core_mask > _MAX_CORE_MASK:
            raise SynapticConfigurationException(
                f"The core mask of {core_mask} is too big "
                f"(maximum {_MAX_CORE_MASK})")
        return self.__add_invalid_entry(
            key_and_mask, core_mask, core_shift, n_neurons, n_colour_bits)

    def __add_invalid_entry(
            self, key_and_mask: BaseKeyAndMask, core_mask: int,
            core_shift: int, n_neurons: int, n_colour_bits: int) -> int:
        """
        Add an entry to the table that doesn't point to anywhere.  Used
        to keep indices in synchronisation between e.g. normal and delay
        entries and between entries on different cores.

        :param key_and_mask:
            a key_and_mask object used as part of describing
            an edge that will require being received to be stored in the
            master pop table; the whole edge will become multiple calls to
            this function
        :param core_mask:
            Mask for the part of the key that identifies the core
        :param core_shift: The shift of the mask to get to the core_mask
        :param n_neurons:
            The number of neurons in each machine vertex (bar the last)
        :param n_colour_bits:
            The number of bits used for colour
        :return: The index of the added entry
        """
        entry = self.__add_entry(
            key_and_mask, core_mask, core_shift, n_neurons, n_colour_bits)
        index = entry.append_invalid()
        self.__n_addresses += 1
        return index

    def __add_entry(
            self, key_and_mask: BaseKeyAndMask, core_mask: int,
            core_shift: int, n_neurons: int,
            n_colour_bits: int) -> _MasterPopEntry:
        if self.__n_addresses >= _MAX_ADDRESS_START:
            raise SynapticConfigurationException(
                f"The table already contains {self.__n_addresses} entries;"
                " adding another is too many")
        if key_and_mask.key not in self.__entries:
            entry = _MasterPopEntry(
                key_and_mask.key, key_and_mask.mask, core_mask, core_shift,
                n_neurons, n_colour_bits)
            self.__entries[key_and_mask.key] = entry
            return entry
        entry = self.__entries[key_and_mask.key]
        if (key_and_mask.mask != entry.mask or
                core_mask != entry.core_mask or
                core_shift != entry.core_shift or
                n_neurons != entry.n_neurons):
            raise SynapticConfigurationException(
                f"Existing entry for key {key_and_mask.key} doesn't match one "
                f"being added: Existing mask: {entry.mask} "
                f"core_mask: {entry.core_mask} core_shift: {entry.core_shift} "
                f"n_neurons: {entry.n_neurons} "
                f"Adding mask: {key_and_mask.mask} core_mask: {core_mask} "
                f"core_shift: {core_shift} n_neurons: {n_neurons}")
        return entry

    def get_pop_table_data(self) -> NDArray[uint32]:
        """
        :returns: The master pop table data as a numpy array.
        """
        # sort entries by key
        entries = sorted(
            self.__entries.values(),
            key=lambda a_entry: a_entry.routing_key)
        n_entries = len(entries)
        data = [numpy.array([n_entries, self.__n_addresses], dtype=uint32)]
        # Generate the table and list as arrays
        pop_table = _make_array(_MasterPopEntryCType, n_entries)
        address_list = _make_array(_AddressListEntryCType, self.__n_addresses)
        start = 0
        for i, entry in enumerate(entries):
            start += entry.write_to_table(pop_table[i], address_list, start)

        # Add the arrays
        data.append(_to_numpy(pop_table))
        data.append(_to_numpy(address_list))
        return numpy.concatenate(data)

    @property
    def max_n_neurons_per_core(self) -> int:
        """
        The maximum number of neurons per core supported when a core-mask
        is > 0.
        """
        return _MAX_N_NEURONS

    @property
    def max_core_mask(self) -> int:
        """
        The maximum core mask supported when n_neurons is > 0; this is the
        maximum number of cores that can be supported in a joined mask.
        """
        return _MAX_CORE_MASK

    @property
    def max_index(self) -> int:
        """
        The maximum index of a synaptic connection.
        """
        return _MAX_ADDRESS_COUNT



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional
from spinn_front_end_common.utility_models import MultiCastCommand
from spinn_front_end_common.utilities.connections import LiveEventConnection
from spynnaker.pyNN.external_devices_models import AbstractEthernetTranslator


class EthernetControlConnection(LiveEventConnection):
    """
    A connection that can translate Ethernet control messages received
    from a Population.
    """
    __slots__ = ("__translators", )

    def __init__(
            self, translator: AbstractEthernetTranslator, label: str,
            live_packet_gather_label: str, local_host: Optional[str] = None,
            local_port: Optional[int] = None):
        """
        :param translator: The translator of multicast to control commands
        :param label: The label of the vertex to attach the translator to
        :param live_packet_gather_label: The label of the LPG vertex that
            this control connection will listen to.
        :param local_host: The optional host to listen on
        :param local_port: The optional port to listen on
        """
        super().__init__(
            live_packet_gather_label, receive_labels=[label],
            local_host=local_host, local_port=local_port)
        self.__translators: Dict[str, AbstractEthernetTranslator] = dict()
        self.__translators[label] = translator
        self.add_receive_no_time_callback(
            label, self._translate, translate_key=False)

    def add_translator(
            self, label: str, translator: AbstractEthernetTranslator) -> None:
        """
        Add another translator that routes via the LPG.

        :param label: The label of the vertex to attach the translator to
        :param translator: The translator of multicast to control commands
        """
        super().add_receive_label(label)
        self.__translators[label] = translator
        self.add_receive_no_time_callback(
            label, self._translate, translate_key=False)

    def _translate(self, label: str, key: int,
                   payload: Optional[int] = None) -> None:
        translator = self.__translators[label]
        if payload is None:
            translator.translate_control_packet(MultiCastCommand(key))
        else:
            translator.translate_control_packet(MultiCastCommand(key, payload))



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.constants import NOTIFY_PORT
from spinn_front_end_common.utilities.database import DatabaseConnection
from spynnaker.pyNN.external_devices_models import AbstractEthernetTranslator


class EthernetCommandConnection(DatabaseConnection):
    """
    A connection that can send commands to a device at the start and end
    of a simulation.
    """
    __slots__ = (
        "__command_containers",
        "__translator")

    def __init__(
            self, translator: AbstractEthernetTranslator,
            command_containers: Optional[Iterable[
                AbstractSendMeMulticastCommandsVertex]] = None,
            local_host: Optional[str] = None,
            local_port: Optional[int] = NOTIFY_PORT):
        """
        :param translator:
            A translator of multicast commands to device commands
        :param command_containers:
            A list of vertices that have commands to be sent at the start
            and end of simulation
        :param local_host:
            The optional host to listen on for the start/resume message
        :param local_port:
            The optional port to listen on for the stop/pause message
        """
        super().__init__(
            start_resume_callback_function=self._start_resume_callback,
            stop_pause_callback_function=self._stop_pause_callback,
            local_host=local_host, local_port=local_port)

        self.__command_containers: List[
            AbstractSendMeMulticastCommandsVertex] = list()
        if command_containers is not None:
            for command_container in command_containers:
                self.add_command_container(command_container)
        self.__translator = translator

    def add_command_container(
            self,
            command_container: AbstractSendMeMulticastCommandsVertex) -> None:
        """
        Add a command container.

        :param command_container:
            A vertex that has commands to be sent at the start and end of
            simulation
        """
        if not isinstance(
                command_container, AbstractSendMeMulticastCommandsVertex):
            raise TypeError(
                "Each command container must be an instance of"
                " AbstractSendMeMulticastCommandsVertex")
        if command_container.timed_commands:
            raise TypeError("Timed commands cannot be handled by this class")
        self.__command_containers.append(command_container)

    def _start_resume_callback(self) -> None:
        # Send commands from each command container
        for command_container in self.__command_containers:
            for command in command_container.start_resume_commands:
                self.__translator.translate_control_packet(command)

    def _stop_pause_callback(self) -> None:
        # Send commands from each command container
        for command_container in self.__command_containers:
            for command in command_container.pause_stop_commands:
                self.__translator.translate_control_packet(command)



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, final, Iterable
from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.exceptions import PacmanConfigurationException
from pacman.model.graphs.common import Slice
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.utilities.algorithm_utilities\
    .partition_algorithm_utilities import get_multidimensional_slices
from pacman.model.graphs.machine.machine_vertex import MachineVertex
from spynnaker.pyNN.models.neuron import PopulationVertex
from .abstract_spynnaker_splitter_delay import AbstractSpynnakerSplitterDelay


class SplitterPopulationVertex(
        AbstractSplitterCommon[PopulationVertex],
        AbstractSpynnakerSplitterDelay, metaclass=AbstractBase):
    """
    Abstract base class of splitters for :py:class:`PopulationVertex`.
    """
    __slots__ = ("__slices", )

    def __init__(self) -> None:
        super().__init__()
        self.__slices: Optional[List[Slice]] = None

    @final
    @overrides(AbstractSplitterCommon.set_governed_app_vertex)
    def set_governed_app_vertex(
            self, app_vertex: PopulationVertex) -> None:
        if not isinstance(app_vertex, PopulationVertex):
            raise PacmanConfigurationException(
                f"The vertex {app_vertex} cannot be supported by the "
                f"{self.__class__.__name__} as the only vertex "
                "supported by this splitter is a PopulationVertex. "
                "Please use the correct splitter for your vertex and try "
                "again.")
        super().set_governed_app_vertex(app_vertex)

    @overrides(AbstractSplitterCommon.reset_called)
    def reset_called(self) -> None:
        self.__slices = None

    @final
    @overrides(AbstractSpynnakerSplitterDelay.max_support_delay)
    def max_support_delay(self) -> int:
        return self.governed_app_vertex.max_delay_steps_incoming

    @overrides(AbstractSpynnakerSplitterDelay.accepts_edges_from_delay_vertex)
    def accepts_edges_from_delay_vertex(self) -> bool:
        return self.governed_app_vertex.allow_delay_extension

    @final
    def _get_fixed_slices(self) -> List[Slice]:
        """
        Get a list of fixed slices from the Application vertex.
        """
        if self.__slices is not None:
            return self.__slices
        self.__slices = get_multidimensional_slices(self.governed_app_vertex)
        return self.__slices

    @abstractmethod
    def machine_vertices_for_recording(
            self, variable_to_record: str) -> Iterable[MachineVertex]:
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary
from spinn_front_end_common.interface.ds import DataType
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractStandardNeuronComponent, ModelParameter)
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView
from .neuron_model import NeuronModel

A = 'a'
B = 'b'
C = 'c'
D = 'd'
V = 'v'
U = 'u'
I_OFFSET = 'i_offset'
TIMESTEP = 'timestep'
NEXT_H = 'next_h'


class NeuronModelIzh(NeuronModel):
    """
    Model of neuron due to Eugene M. Izhikevich et al.
    """
    __slots__ = (
        "__a", "__b", "__c", "__d",
        "__v_init", "__u_init", "__i_offset")

    def __init__(
            self, a: ModelParameter, b: ModelParameter, c: ModelParameter,
            d: ModelParameter, v_init: ModelParameter, u_init: ModelParameter,
            i_offset: ModelParameter):
        """
        :param a: :math:`a`
        :param b: :math:`b`
        :param c: :math:`c`
        :param d: :math:`d`
        :param v_init: :math:`v_{init}`
        :param u_init: :math:`u_{init}`
        :param i_offset: :math:`I_{offset}`
        """
        super().__init__(
            [Struct([
                (DataType.S1615, A),
                (DataType.S1615, B),
                (DataType.S1615, C),
                (DataType.S1615, D),
                (DataType.S1615, V),
                (DataType.S1615, U),
                (DataType.S1615, I_OFFSET),
                (DataType.S1615, TIMESTEP),
                (DataType.S1615, NEXT_H)])],
            {A: "ms", B: "ms", C: "mV", D: "mV/ms", V: "mV", U: "mV/ms",
             I_OFFSET: "nA"})
        self.__a = a
        self.__b = b
        self.__c = c
        self.__d = d
        self.__i_offset = i_offset
        self.__v_init = v_init
        self.__u_init = u_init

    @overrides(AbstractStandardNeuronComponent.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[A] = self._convert(self.__a)
        parameters[B] = self._convert(self.__b)
        parameters[C] = self._convert(self.__c)
        parameters[D] = self._convert(self.__d)
        parameters[I_OFFSET] = self._convert(self.__i_offset)
        parameters[TIMESTEP] = SpynnakerDataView.get_simulation_time_step_ms()

    @overrides(AbstractStandardNeuronComponent.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[V] = self._convert(self.__v_init)
        state_variables[U] = self._convert(self.__u_init)
        state_variables[NEXT_H] = (
            SpynnakerDataView.get_simulation_time_step_ms())

    @property
    def a(self) -> ModelParameter:
        """
        Settable model parameter: :math:`a`
        """
        return self.__a

    @property
    def b(self) -> ModelParameter:
        """
        Settable model parameter: :math:`b`
        """
        return self.__b

    @property
    def c(self) -> ModelParameter:
        """
        Settable model parameter: :math:`c`
        """
        return self.__c

    @property
    def d(self) -> ModelParameter:
        """
        Settable model parameter: :math:`d`
        """
        return self.__d

    @property
    def i_offset(self) -> ModelParameter:
        """
        Settable model parameter: :math:`I_{offset}`
        """
        return self.__i_offset

    @property
    def v_init(self) -> ModelParameter:
        """
        Settable model parameter: :math:`v_{init}`
        """
        return self.__v_init

    @property
    def u_init(self) -> ModelParameter:
        """
        Settable model parameter: :math:`u_{init}`
        """
        return self.__u_init



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, Tuple
from spinn_utilities.overrides import overrides
from spynnaker.pyNN.models.neuron import PopulationVertex
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from spynnaker.pyNN.utilities.constants import POP_TABLE_MAX_ROW_LENGTH
if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.implementations import AbstractNeuronImpl
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        SplitterPopulationVertex)

# The maximum atoms per core is the master population table row length to
# make it easier when all-to-all-connector is used
DEFAULT_MAX_ATOMS_PER_CORE = POP_TABLE_MAX_ROW_LENGTH

_population_parameters: Dict[str, Any] = {
    "spikes_per_second": None, "ring_buffer_sigma": None,
    "max_expected_summed_weight": None,
    "incoming_spike_buffer_size": None, "drop_late_spikes": None,
    "splitter": None, "seed": None, "n_colour_bits": None,
    "n_synapse_cores": None, "allow_delay_extensions": None,
    "neurons_per_core": None,
}


class AbstractPyNNNeuronModel(AbstractPyNNModel):
    """
    API for a PyNN Neuron Model
    """
    __slots__ = ("__model", )

    # The number of synapse cores for PyNN models that use PopulationVertex
    # or None to determine based on time-step
    _n_synapse_cores: Dict[type, Optional[int]] = {}

    # Whether to allow delay extensions when using PyNN models that use
    # PopulationVertex
    _allow_delay_extensions: Dict[type, bool] = {}

    #: Population parameters for neuron models.
    default_population_parameters = _population_parameters

    @classmethod
    def set_model_n_synapse_cores(cls, n_synapse_cores: Optional[int]) -> None:
        """
        Set the number of synapse cores for a model.

        :param n_synapse_cores:
            The number of synapse cores; 0 to force combined cores, and None to
            allow the system to choose
        """
        cls.verify_may_set(param="n_synapse_cores")
        cls._n_synapse_cores[cls] = n_synapse_cores

    @classmethod
    def get_model_n_synapse_cores(cls) -> Optional[int]:
        """
        :returns: The number of synapse cores for the model.
        """
        return cls._n_synapse_cores.get(cls, None)

    @classmethod
    def set_model_allow_delay_extensions(cls, allow: bool) -> None:
        """
        Set whether to allow delay extensions for a model.

        :param allow: Whether to allow delay extensions
        """
        cls.verify_may_set(param="allow_delay_extensions")
        cls._allow_delay_extensions[cls] = allow

    @classmethod
    def get_model_allow_delay_extensions(cls) -> bool:
        """
        Get whether to allow delay extensions for the model.

        :returns: True unless the model does not allow delay extensions
        """
        return cls._allow_delay_extensions.get(cls, True)

    @classmethod
    @overrides(AbstractPyNNModel.reset_all)
    def reset_all(cls) -> None:
        super().reset_all()
        AbstractPyNNNeuronModel._n_synapse_cores.clear()
        AbstractPyNNNeuronModel._allow_delay_extensions.clear()

    def __init__(self, model: AbstractNeuronImpl):
        """
        :param model: The model implementation
        """
        self.__model = model

    @property
    def _model(self) -> AbstractNeuronImpl:
        return self.__model

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *,
            spikes_per_second: Optional[float] = None,
            ring_buffer_sigma: Optional[float] = None,
            max_expected_summed_weight: Optional[List[float]] = None,
            incoming_spike_buffer_size: Optional[int] = None,
            drop_late_spikes: Optional[bool] = None,
            splitter: Optional[SplitterPopulationVertex] = None,
            seed: Optional[int] = None,
            n_colour_bits: Optional[int] = None,
            neurons_per_core: Optional[Union[int, Tuple[int, ...]]] = None,
            n_synapse_cores: Optional[int] = None,
            allow_delay_extensions: Optional[bool] = None) -> PopulationVertex:
        """
        :param spikes_per_second:
        :param ring_buffer_sigma:
        :param incoming_spike_buffer_size:
        :param drop_late_spikes:
        :param splitter:
        :param seed:
        :param n_colour_bits:
        """
        if neurons_per_core is None:
            neurons_per_core = \
                self.get_model_max_atoms_per_dimension_per_core()
        if n_synapse_cores is None:
            n_synapse_cores = self.get_model_n_synapse_cores()
        if allow_delay_extensions is None:
            allow_delay_extensions = self.get_model_allow_delay_extensions()
        return PopulationVertex(
            n_neurons=n_neurons, label=label,
            max_atoms_per_core=neurons_per_core,
            n_synapse_cores=n_synapse_cores,
            allow_delay_extensions=allow_delay_extensions,
            spikes_per_second=spikes_per_second,
            ring_buffer_sigma=ring_buffer_sigma,
            max_expected_summed_weight=max_expected_summed_weight,
            incoming_spike_buffer_size=incoming_spike_buffer_size,
            neuron_impl=self.__model, pynn_model=self,
            drop_late_spikes=drop_late_spikes or False,
            splitter=splitter, seed=seed, n_colour_bits=n_colour_bits)

    @property
    @overrides(AbstractPyNNModel.name)
    def name(self) -> str:
        return self.__model.model_name



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
from threading import Thread
from typing import Callable, Dict, Final, Iterable, List, Optional, Set, Tuple
from typing_extensions import TypeAlias
from spinn_utilities.log import FormatAdapter
from spinnman.connections import ConnectionListener
from spinnman.connections.udp_packet_connections import UDPConnection
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.database import DatabaseConnection
from spinn_front_end_common.utilities.database import DatabaseReader

_EVENT: Final['TypeAlias'] = Callable[[str, List[int]], None]
_INIT: Final['TypeAlias'] = Callable[[str, int, float, float], None]
_START_STOP: Final['TypeAlias'] = Callable[
    [str, 'SPIFLiveSpikesConnection'], None]
logger = FormatAdapter(logging.getLogger(__name__))

_TWO_SKIP = struct.Struct("<2x")

_THREE_INTS = struct.Struct("<3I")

_ONE_INT = struct.Struct("<I")

# The port that SPIF listens on by default
_DEFAULT_SPIF_PORT = 3332

# The maximum number of events in each packet by default
_EVENTS_PER_PACKET = 32

# The maximum time between packets in microseconds by default
_US_PER_PACKET = 500

# SPIF message to start sending
_SPIF_OUTPUT_START = 0x5ec00000

# SPIF message to stop sending
_SPIF_OUTPUT_STOP = 0x5ec10000

# SPIF message to set packet send time (time is added to this in microseconds)
_SPIF_OUTPUT_SET_TICK = 0x5ec20000

# SPIF message to set packet size (size is added to this in bytes)
_SPIF_OUTPUT_SET_LEN = 0x5ec40000


class SPIFLiveSpikesConnection(DatabaseConnection):
    """
    A connection for receiving live spikes from SPIF.
    """
    # TODO: define SPIF
    __slots__ = (
        "__error_keys",
        "__init_callbacks",
        "__key_to_atom_id_and_label",
        "__live_event_callbacks",
        "__pause_stop_callbacks",
        "__receive_labels",
        "__receiver_connection",
        "__receiver_listener",
        "__start_resume_callbacks",
        "__spif_host",
        "__spif_port",
        "__spif_packet_size",
        "__spif_packet_time_us")

    def __init__(self, receive_labels: Optional[Iterable[str]],
                 spif_host: str, spif_port: int = _DEFAULT_SPIF_PORT,
                 events_per_packet: int = _EVENTS_PER_PACKET,
                 time_per_packet: int = _US_PER_PACKET,
                 local_host: Optional[str] = None,
                 local_port: Optional[int] = None):
        """
        :param receive_labels:
            Labels of vertices from which live events will be received.
        :param spif_host: The location of the SPIF board receiving packets
        :param spif_port: The port of the SPIF board (default 3332)
        :param events_per_packet:
            The maximum number of events in each packet.  SPIF will be
            configured to send a packet as soon as it reaches this size if not
            before (default is 32)
        :param time_per_packet:
            The maximum time between sending non-empty packets.  SPIF will be
            configured to send a packet that isn't empty after this many
            microseconds (default is 500)
        :param local_host:
            Optional specification of the local hostname or IP address of the
            interface to listen on
        :param local_port:
            Optional specification of the local port to listen on. Must match
            the port that the toolchain will send the notification on (19999
            by default)
        """
        super().__init__(
            self.__do_start_resume, self.__do_stop_pause,
            local_host=local_host, local_port=local_port)

        self.add_database_callback(self.__read_database_callback)

        self.__receive_labels = (
            list(receive_labels) if receive_labels is not None else [])

        self.__spif_host = spif_host
        self.__spif_port = spif_port
        self.__spif_packet_size = events_per_packet * BYTES_PER_WORD
        self.__spif_packet_time_us = time_per_packet
        self.__key_to_atom_id_and_label: Dict[int, Tuple[int, int]] = dict()
        self.__live_event_callbacks: List[List[Tuple[_EVENT, bool]]] = list()
        self.__start_resume_callbacks: Dict[str, List[_START_STOP]] = dict()
        self.__pause_stop_callbacks: Dict[str, List[_START_STOP]] = dict()
        self.__init_callbacks: Dict[str, List[_INIT]] = dict()
        if receive_labels is not None:
            for label in receive_labels:
                self.__live_event_callbacks.append(list())
                self.__start_resume_callbacks[label] = list()
                self.__pause_stop_callbacks[label] = list()
                self.__init_callbacks[label] = list()
        self.__receiver_listener: Optional[ConnectionListener[bytes]] = None
        self.__receiver_connection: Optional[UDPConnection] = None
        self.__error_keys: Set[int] = set()

    def add_receive_label(self, label: str) -> None:
        """
        Adds this labels to the callbacks.

        :param label:
        """
        if label not in self.__receive_labels:
            self.__receive_labels.append(label)
            self.__live_event_callbacks.append(list())
        if label not in self.__start_resume_callbacks:
            self.__start_resume_callbacks[label] = list()
            self.__pause_stop_callbacks[label] = list()
            self.__init_callbacks[label] = list()

    def add_init_callback(self, label: str, init_callback: _INIT) -> None:
        """
        Add a callback to be called to initialise a vertex.

        :param label:
            The label of the vertex to be notified about. Must be one of the
            vertices listed in the constructor
        :param init_callback: A function to be called to initialise the
            vertex. This should take as parameters the label of the vertex,
            the number of neurons in the population, the run time of the
            simulation in milliseconds, and the simulation timestep in
            milliseconds
        """
        self.__init_callbacks[label].append(init_callback)

    def add_receive_callback(
            self, label: str, live_event_callback: _EVENT,
            translate_key: bool = True) -> None:
        """
        Add a callback for the reception of live events from a vertex.

        :param label: The label of the vertex to be notified about.
            Must be one of the vertices listed in the constructor
        :param live_event_callback: A function to be called when events are
            received. This should take as parameters the label of the vertex,
            and an array-like of atom IDs.
        :param translate_key:
            True if the key is to be converted to an atom ID, False if the
            key should stay a key
        """
        label_id = self.__receive_labels.index(label)
        logger.info(
            "Receive callback {} registered to label {}",
            live_event_callback, label)
        self.__live_event_callbacks[label_id].append(
            (live_event_callback, translate_key))

    def add_start_resume_callback(
            self, label: str, start_resume_callback: _START_STOP) -> None:
        """
        Add a callback for the start and resume state of the simulation.

        :param label: the label of the function to be sent
        :param start_resume_callback: A function to be called when the start
            or resume message has been received. This function should take
            the label of the referenced vertex, and an instance of this
            class, which can be used to send events.
        """
        self.__start_resume_callbacks[label].append(start_resume_callback)

    def add_pause_stop_callback(
            self, label: str, pause_stop_callback: _START_STOP) -> None:
        """
        Add a callback for the pause and stop state of the simulation.

        :param label: the label of the function to be sent
        :param pause_stop_callback: A function to be called when the pause
            or stop message has been received. This function should take the
            label of the referenced  vertex, and an instance of this class,
            which can be used to send events.
        """
        self.__pause_stop_callbacks[label].append(pause_stop_callback)

    def __read_database_callback(self, db_reader: DatabaseReader) -> None:
        self.__handle_possible_rerun_state()

        vertex_sizes: Dict[str, int] = dict()
        run_time_ms = db_reader.get_configuration_parameter_value(
            "runtime") or 0.0
        machine_timestep_ms = (
            db_reader.get_configuration_parameter_value("machine_time_step")
            or 0.0) / 1000.0

        if self.__receive_labels:
            self.__init_receivers(db_reader, vertex_sizes)

        for label, vertex_size in vertex_sizes.items():
            for init_callback in self.__init_callbacks[label]:
                init_callback(
                    label, vertex_size, run_time_ms, machine_timestep_ms)

    def __init_receivers(
            self, db: DatabaseReader, vertex_sizes: Dict[str, int]) -> None:
        # Set up a single connection for receive
        if self.__receiver_connection is None:
            self.__receiver_connection = UDPConnection(
                remote_host=self.__spif_host, remote_port=self.__spif_port)
        for label_id, label in enumerate(self.__receive_labels):
            key_to_atom_id = db.get_key_to_atom_id_mapping(label)
            for key, atom_id in key_to_atom_id.items():
                self.__key_to_atom_id_and_label[key] = (atom_id, label_id)
            vertex_sizes[label] = len(key_to_atom_id)

        # Last of all, set up the listener for packets
        # NOTE: Has to be done last as otherwise will receive SCP messages
        # sent above!
        if self.__receiver_listener is None:
            self.__receiver_listener = ConnectionListener(
                self.__receiver_connection)
            self.__receiver_listener.add_callback(self.__do_receive_packet)
            self.__receiver_listener.start()

    def __handle_possible_rerun_state(self) -> None:
        # reset from possible previous calls
        if self.__receiver_listener is not None:
            self.__receiver_listener.close()
            self.__receiver_listener = None
        if self.__receiver_connection is not None:
            self.__receiver_connection.close()
            self.__receiver_connection = None

    def __launch_thread(
            self, kind: str, label: str, callback: _START_STOP) -> None:
        thread = Thread(target=callback, args=(label, self), name=(
            f"{kind} callback thread for live_event_connection "
            f"{self._local_port}:{self._local_ip_address}"))
        thread.start()

    def __do_start_resume(self) -> None:
        if not self.__receiver_connection:
            return
        # Send SPIF configuration
        self.__receiver_connection.send(_THREE_INTS.pack(
            _SPIF_OUTPUT_SET_LEN + self.__spif_packet_size,
            _SPIF_OUTPUT_SET_TICK + self.__spif_packet_time_us,
            _SPIF_OUTPUT_START))
        for label, callbacks in self.__start_resume_callbacks.items():
            for callback in callbacks:
                self.__launch_thread("start_resume", label, callback)

    def __do_stop_pause(self) -> None:
        if not self.__receiver_connection:
            return
        # Stop SPIF output
        self.__receiver_connection.send(_ONE_INT.pack(_SPIF_OUTPUT_STOP))
        for label, callbacks in self.__pause_stop_callbacks.items():
            for callback in callbacks:
                self.__launch_thread("pause_stop", label, callback)

    def __do_receive_packet(self, packet: bytes) -> None:
        # pylint: disable=broad-except
        logger.debug("Received packet")
        try:
            self.__handle_packet(packet)
        except Exception:
            logger.warning("problem handling received packet", exc_info=True)

    def __handle_packet(self, packet: bytes) -> None:
        key_labels: Dict[int, List[int]] = dict()
        atoms_labels: Dict[int, List[int]] = dict()
        n_events = len(packet) // BYTES_PER_WORD
        events = struct.unpack(f"<{n_events}I", packet)
        for key in events:
            if key in self.__key_to_atom_id_and_label:
                atom_id, label_id = self.__key_to_atom_id_and_label[key]
                if label_id not in key_labels:
                    key_labels[label_id] = list()
                    atoms_labels[label_id] = list()
                key_labels[label_id].append(key)
                atoms_labels[label_id].append(atom_id)
            else:
                self.__handle_unknown_key(key)

        for label_id in key_labels:
            label = self.__receive_labels[label_id]
            for c_back, use_atom in self.__live_event_callbacks[label_id]:
                if use_atom:
                    c_back(label, atoms_labels[label_id])
                else:
                    c_back(label, key_labels[label_id])

    def __handle_unknown_key(self, key: int) -> None:
        if key not in self.__error_keys:
            self.__error_keys.add(key)
            logger.warning("Received unexpected key {}", key)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy
from numpy import integer, uint8, uint32
from numpy.typing import NDArray
from pyNN.random import RandomDistribution
from typing_extensions import TypeAlias

from spinn_utilities.helpful_functions import is_singleton
from spinn_utilities.ranged.abstract_list import AbstractList
from spinn_utilities.ranged.range_dictionary import RangeDictionary

from pacman.model.graphs.common import Slice

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.utilities.utility_calls import convert_to
from spynnaker.pyNN.models.common.param_generator_data import (
    get_generator_type, param_generator_id, param_generator_params,
    type_has_generator)

#: The type of values used populate structure instances
ValueMap: TypeAlias = Mapping[str, Union[int, float, AbstractList[float]]]


REPEAT_PER_NEURON_FLAG = 0xFFFFFFFF


class StructRepeat(Enum):
    """
    How a structure repeats, or not, in memory.
    """
    #: Indicates a single global struct
    GLOBAL = 0

    #: Indicates a struct that repeats per neuron
    PER_NEURON = 1


class Struct(object):
    """
    Represents a C code structure.
    """
    __slots__ = (
        "__fields",
        "__repeat_type",
        "__default_values")

    def __init__(
            self, fields: Sequence[Tuple[DataType, str]],
            repeat_type: StructRepeat = StructRepeat.PER_NEURON,
            default_values: Optional[Dict[str, Union[int, float]]] = None):
        """
        :param fields:
            The types and names of the fields, ordered as they appear in the
            structure.
        :param repeat_type: How the structure repeats
        :param default_values:
            Dict of field name -> value to use when values doesn't contain the
            field
        """
        self.__fields = fields
        self.__repeat_type = repeat_type
        self.__default_values = default_values or dict()

    @property
    def fields(self) -> Sequence[Tuple[DataType, str]]:
        """
        The types and names of the fields, ordered as they appear in the
        structure.
        """
        return self.__fields

    @property
    def repeat_type(self) -> StructRepeat:
        """
        How the structure repeats.
        """
        return self.__repeat_type

    @property
    def numpy_dtype(self) -> numpy.dtype:
        """
        The numpy data type of the structure.
        """
        return numpy.dtype(
            [(name, numpy.dtype(data_type.struct_encoding))
             for data_type, name in self.__fields],
            align=True)

    def get_size_in_whole_words(self, array_size: int = 1) -> int:
        """
        :param array_size: The number of elements in an array of structures
        :returns: The size of the structure in whole words in an array
            of given size (default 1 item).
        """
        datatype = self.numpy_dtype
        size_in_bytes = array_size * datatype.itemsize
        return (size_in_bytes + (BYTES_PER_WORD - 1)) // BYTES_PER_WORD

    def get_data(self, values: ValueMap,
                 vertex_slice: Optional[Slice] = None) -> NDArray[uint32]:
        """
        :param values: The values to fill in the data with
        :param vertex_slice:
            The vertex slice to get the data for, or `None` if the structure is
            global.
        :returns: The numpy array of uint32 of data for the given values
        """
        n_items = 1
        if vertex_slice is None:
            if self.__repeat_type != StructRepeat.GLOBAL:
                raise ValueError(
                    "Repeating structures must specify a vertex_slice")
        elif self.__repeat_type == StructRepeat.GLOBAL:
            raise ValueError("Global Structures do not have a slice")
        else:
            n_items = vertex_slice.n_atoms

        # Create an array to store values in
        data = numpy.zeros(n_items, dtype=self.numpy_dtype)

        if not self.__fields:
            return data.view(uint32)

        # Go through and get the values and put them in the array
        for data_type, name in self.__fields:
            if name in values:
                all_vals = values[name]
                if is_singleton(all_vals):
                    # If there is just one value for everything, use it
                    # everywhere
                    data[name] = convert_to(all_vals, data_type)
                    continue
                assert isinstance(all_vals, AbstractList)
                if self.__repeat_type == StructRepeat.GLOBAL:
                    # If there is a ranged list for global struct,
                    # we might need to read a single value
                    data[name] = convert_to(
                        all_vals.get_single_value_all(), data_type)
                else:
                    assert vertex_slice is not None
                    self.__get_data_for_slice(
                        data, all_vals, name, data_type, vertex_slice)
            else:
                # If there is only a default value, get that and use it
                # everywhere
                value = self.__default_values[name]
                data_value = convert_to(value, data_type)
                data[name] = data_value

        # Pad to whole number of uint32s
        overflow = (n_items * self.numpy_dtype.itemsize) % BYTES_PER_WORD
        if overflow != 0:
            data = numpy.pad(
                data.view(uint8), (0, BYTES_PER_WORD - overflow), "constant")

        return data.view(uint32)

    def __get_data_for_slice(
            self, data: NDArray, all_vals: AbstractList[float], name: str,
            data_type: DataType, vertex_slice: Slice) -> None:
        """
        Get the data for a single value from a vertex slice.
        """
        # If there is a list of values, convert it
        ids = vertex_slice.get_raster_ids()
        data_pos = 0
        for start, stop, value in all_vals.iter_ranges_by_ids(ids):
            # Get the values and convert to the correct data type
            n_values = stop - start
            if isinstance(value, RandomDistribution):
                r_vals = value.next(n_values)
                data[name][data_pos:data_pos + n_values] = [
                    convert_to(v, data_type) for v in r_vals]
            else:
                data[name][data_pos:data_pos + n_values] = convert_to(
                    value, data_type)

            data_pos += n_values

    def get_generator_data(
            self, values: ValueMap,
            vertex_slice: Optional[Slice] = None) -> NDArray[uint32]:
        """
        :param values: The values to fill in the data with
        :param vertex_slice:
            The vertex slice or `None` for a structure with repeat_type global,
            or where a single value repeats for every neuron.  If this is not
            the case and vertex_slice is `None`, an error will be raised!
        :returns: A numpy array of uint32 of data to generate the given values.
        """
        # Define n_repeats, which is either the total number of neurons
        # or a flag to indicate that the data repeats for each neuron
        if vertex_slice is None:
            if self.__repeat_type == StructRepeat.GLOBAL:
                n_repeats = 1
            else:
                n_repeats = REPEAT_PER_NEURON_FLAG
        else:
            if self.__repeat_type == StructRepeat.GLOBAL:
                raise ValueError(
                    "Global Structures cannot repeat more than once")
            n_repeats = vertex_slice.n_atoms

        # Start with bytes per repeat, n_repeats (from above),
        # total size of data written (0 as filled in later),
        # and number of fields in struct
        data = [self.numpy_dtype.itemsize, n_repeats, 0, len(self.__fields)]
        gen_data: List[NDArray[uint32]] = list()

        # Go through all values and add in generator data for each
        for data_type, name in self.__fields:
            # Store the writer type based on the data type
            data.append(get_generator_type(data_type))

            # We want the data generated "per neuron" regardless of how many -
            # there must be a single value for this to work
            if vertex_slice is None:
                self.__gen_data_one_for_all(data, gen_data, values, name,
                                            n_repeats)

            # If we know the array size, the values can vary per neuron
            else:
                self.__gen_data_for_slice(
                    data, gen_data, values, name, vertex_slice)

        # Update with size *before* adding generator parameters
        data[2] = len(data) * BYTES_PER_WORD

        # Add the generator parameters after the rest of the data
        all_data = [numpy.array(data, dtype=uint32)]
        all_data.extend(gen_data)

        # Make it one
        return numpy.concatenate(all_data)

    def __gen_data_one_for_all(
            self, data: List[int], gen_data: List[NDArray[uint32]],
            values: ValueMap, name: str, n_repeats: int) -> None:
        """
        Generate data with a single value for all neurons.
        """
        # How many sub-sets of repeats there are (1 in this case as
        # that one sub-set covers all neurons)
        data.append(1)

        # How many times to repeat the next bit
        data.append(n_repeats)

        # Get the value to write, of which there can only be one
        # (or else there will be an error here ;)
        value: Any
        if name in values:
            value = values[name]
            if not is_singleton(value):
                value = value.get_single_value_all()
        else:
            value = self.__default_values[name]

        # Write the id of the generator of the parameter
        data.append(param_generator_id(value))

        # Add any parameters required to generate the values
        gen_data.append(param_generator_params(value))

    def __gen_data_for_slice(
            self, data: List[int], gen_data: List[NDArray[uint32]],
            values: ValueMap, name: str, vertex_slice: Slice) -> None:
        """
        Generate data with different values for each neuron.
        """
        # If we have a range list for the value, generate for the range
        if name in values:
            vals = values[name]

            if is_singleton(vals):
                # If there is a single value, we can just use that
                # on all atoms
                data.append(1)
                data.append(vertex_slice.n_atoms)
                data.append(param_generator_id(vals))
                gen_data.append(param_generator_params(vals))
            else:
                assert isinstance(vals, AbstractList)
                # Store where to update with the number of items and
                # set to 0 to start
                n_items_index = len(data)
                data.append(0)
                n_items = 0

                # Go through and get the data for each value
                ids = vertex_slice.get_raster_ids()
                for start, stop, value in vals.iter_ranges_by_ids(ids):
                    n_items += 1
                    # This is the metadata
                    data.append(stop - start)
                    data.append(param_generator_id(value))
                    # This data goes after *all* the metadata
                    gen_data.append(param_generator_params(value))
                data[n_items_index] = n_items
        else:
            # Just a single value for all neurons from defaults
            value = self.__default_values[name]
            data.append(1)
            data.append(vertex_slice.n_atoms)
            data.append(param_generator_id(value))
            gen_data.append(param_generator_params(value))

    @property
    def is_generatable(self) -> bool:
        """
        Whether the data inside could be generated on machine.
        """
        return all(type_has_generator(data_type)
                   for data_type, _name in self.__fields)

    def read_data(
            self, data: bytes, values: RangeDictionary, data_offset: int = 0,
            vertex_slice: Optional[Slice] = None) -> None:
        """
        Read a byte string of data and write to values.

        :param data: The data to be read
        :param values: The values to update with the read data
        :param data_offset:
            Index of the byte at the start of the valid data.
        :param vertex_slice: Slice to read data for or None for all
        """
        n_items = 1
        ids: NDArray[integer] = numpy.zeros([0], dtype=uint32)
        if vertex_slice is None:
            if self.__repeat_type != StructRepeat.GLOBAL:
                raise ValueError(
                    "Repeating structures must specify an array size")
        elif self.__repeat_type == StructRepeat.GLOBAL:
            raise ValueError("Global Structures do not have a slice")
        else:
            n_items = vertex_slice.n_atoms
            ids = vertex_slice.get_raster_ids()

        if not self.__fields:
            return

        # Read in the data values
        numpy_data = numpy.frombuffer(
            data, offset=data_offset, dtype=self.numpy_dtype, count=n_items)

        for data_type, name in self.fields:
            # Ignore fields that can't be set
            if name in values:
                # Get the data to set for this item
                value = data_type.decode_numpy_array(numpy_data[name])
                if self.__repeat_type == StructRepeat.GLOBAL:
                    values[name] = value[0]
                else:
                    values[name].set_value_by_ids(ids, value)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, cast, TYPE_CHECKING
from spinn_utilities.overrides import overrides
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM
from pacman.model.placements import Placement
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.ds import DataSpecificationGenerator
from .population_synapses_machine_vertex_common import (
    PopulationSynapsesMachineVertexCommon)
from .population_machine_synapses_provenance import (
    PopulationMachineSynapsesProvenance)
if TYPE_CHECKING:
    from .population_vertex import PopulationVertex
    from spynnaker.pyNN.models.neuron.synaptic_matrices import (
        SynapseRegionReferences)


class PopulationSynapsesMachineVertexShared(
        PopulationSynapsesMachineVertexCommon,
        PopulationMachineSynapsesProvenance,
        AbstractGeneratesDataSpecification):
    """
    A machine vertex for PyNN Populations.
    """

    __slots__ = ("__synapse_references", )

    def __init__(
            self, sdram: AbstractSDRAM, label: str,
            app_vertex: PopulationVertex, vertex_slice: Slice,
            synapse_references: SynapseRegionReferences):
        """
        :param sdram: The SDRAM used by the vertex
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        """
        super().__init__(sdram, label, app_vertex, vertex_slice)
        self.__synapse_references = synapse_references

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        rec_regions = self._pop_vertex.synapse_recorder.get_region_sizes(
            self.vertex_slice)
        self._write_common_data_spec(spec, rec_regions)
        # Write references to shared regions
        for reg, ref in zip(self.SYNAPSE_REGIONS, self.__synapse_references):
            if ref is not None:
                # Ignore a region if there's no target reference
                spec.reference_memory_region(cast(int, reg), ref)

        # Write information about SDRAM
        self._write_sdram_edge_spec(spec)

        # Write information about keys
        self._write_key_spec(spec)

        # End the writing of this specification:
        spec.end_specification()

    @overrides(PopulationSynapsesMachineVertexCommon._parse_synapse_provenance)
    def _parse_synapse_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        PopulationMachineSynapsesProvenance._parse_synapse_provenance(
            self, label, x, y, p, provenance_data)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.utility_models import MultiCastCommand


class AbstractEthernetTranslator(object, metaclass=AbstractBase):
    """
    A module that can translate packets received over Ethernet into
    control of an external device.
    """

    __slots__ = ()

    @abstractmethod
    def translate_control_packet(
            self, multicast_packet: MultiCastCommand) -> None:
        """
        Translate a multicast packet received over Ethernet and send
        appropriate messages to the external device.

        :param multicast_packet: A received multicast packet
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import os
from typing import Any, Collection, Optional, Union, cast

from lazyarray import __version__ as lazyarray_version

from neo import __version__ as neo_version
from quantities import __version__ as quantities_version
from pyNN.common import control as pynn_control
from pyNN import __version__ as pynn_version
from typing_extensions import Literal, Never


from spinn_utilities.log import FormatAdapter
from spinn_utilities.config_holder import (
    get_config_bool, get_config_str_or_none)
from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.abstract_spinnaker_base import (
    AbstractSpinnakerBase)
from spinn_front_end_common.interface.provenance import (
    FecTimer, GlobalProvenance, TimerCategory, TimerWork)
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker import _version
from spynnaker.pyNN import model_binaries
from spynnaker.pyNN.config_setup import setup_configs
from spynnaker.pyNN.models.recorder import Recorder
from spynnaker.pyNN.models.neuron import AbstractPyNNNeuronModel
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.data.spynnaker_data_writer import SpynnakerDataWriter
from spynnaker.pyNN.extra_algorithms import (
    delay_support_adder, neuron_expander, synapse_expander,
    redundant_packet_count_report,
    spynnaker_neuron_graph_network_specification_report)
from spynnaker.pyNN.extra_algorithms.connection_holder_finisher import (
    finish_connection_holders)
from spynnaker.pyNN.extra_algorithms.splitter_components import (
    spynnaker_splitter_selector)
from spynnaker.pyNN.utilities.neo_buffer_database import NeoBufferDatabase


logger = FormatAdapter(logging.getLogger(__name__))


class SpiNNaker(AbstractSpinnakerBase, pynn_control.BaseState):
    """
    Main interface for the sPyNNaker implementation of PyNN 0.8/0.9.
    """

    __slots__ = ("__recorders", )

    def __init__(
            self, time_scale_factor: Optional[int],
            min_delay: Union[float, None, Literal["auto"]],
            n_chips_required: Optional[int] = None,
            n_boards_required: Optional[int] = None,
            timestep: Optional[float] = 0.1):
        """
        :param time_scale_factor:
            multiplicative factor to the machine time step
            (does not affect the neuron models accuracy)
        :param min_delay:
        :param n_chips_required:
            Deprecated! Use n_boards_required instead.
            Must be `None` if n_boards_required specified.
        :param n_boards_required:
            if you need to be allocated a machine (for spalloc) before
            building your graph, then fill this in with a general idea of
            the number of boards you need so that the spalloc system can
            allocate you a machine big enough for your needs.
        :param timestep:
            the time step of the simulations in microseconds;
            if `None` the cfg value is used
        """
        # change min delay auto to be the min delay supported by simulator
        if min_delay == "auto":
            min_delay = timestep

        # pynn demanded objects
        self.__recorders: Collection[Recorder] = set()

        # main pynn interface inheritance
        pynn_control.BaseState.__init__(self)

        # SpiNNaker setup
        setup_configs()

        # add model binaries
        # called before super.init as that logs the paths
        SpynnakerDataView.register_binary_search_path(
            os.path.dirname(model_binaries.__file__))

        super().__init__(SpynnakerDataWriter)

        self.__writer.set_n_required(n_boards_required, n_chips_required)
        # set up machine targeted data
        self._set_up_timings(timestep, min_delay, time_scale_factor)

        with GlobalProvenance() as db:
            db.insert_version("sPyNNaker_version", _version.__version__)
            db.insert_version("pyNN_version", pynn_version)
            db.insert_version("quantities_version", quantities_version)
            db.insert_version("neo_version", neo_version)
            db.insert_version("lazyarray_version", lazyarray_version)

        # Clears all previously added ceiling on the number of neurons per
        # core, the number of synapse cores and allowing of delay extensions
        AbstractPyNNNeuronModel.reset_all()

    @property
    def __writer(self) -> SpynnakerDataWriter:
        return cast(SpynnakerDataWriter, self._data_writer)

    def _clear_and_run(self, run_time: Optional[float],
                       sync_time: float = 0.0) -> None:
        """
        Clears the projections and Run the model created.

        :param run_time: the time (in milliseconds) to run the simulation for
        :param sync_time:
            If not 0, this specifies that the simulation should pause after
            this duration.  The continue_simulation() method must then be
            called for the simulation to continue.
        """
        # sPyNNaker specific algorithms to do before starting a run
        self.__flush_post_vertex_caches()

        super(SpiNNaker, self).run(run_time, sync_time)

        # PyNNaker specific algorithms to do after finishing a run
        self.__flush_post_vertex_caches()

    def __flush_post_vertex_caches(self) -> None:
        # pylint: disable=protected-access
        for projection in self.__writer.iterate_projections():
            projection._clear_cache()

    def run(self, run_time: Optional[float], sync_time: float = 0.0) -> None:
        """
        Run the simulation for a span of simulation time.

        :param run_time: the time to run for, in milliseconds
        :param sync_time:
            If not 0, this specifies that the simulation should pause after
            this duration.  The continue_simulation() method must then be
            called for the simulation to continue.
        """
        self._clear_and_run(run_time, sync_time)

    def run_until(self, tstop: float) -> None:
        """
        Run the simulation until the given simulation time.

        :param tstop: when to run until in milliseconds
        """
        # Build data
        self._clear_and_run(tstop - self.t)

    def clear(self) -> None:
        """
        Clear the current recordings and reset the simulation.
        """
        self.recorders = set()
        self.reset()

        # Stop any currently running SpiNNaker application
        self.stop()

    @property
    def state(self) -> 'SpiNNaker':
        """
        In SpyNNaker the state and SpiNNaker object are the same
        """
        return self

    @property
    def mpi_rank(self) -> int:
        """
        The MPI rank of the simulator.

        .. note::
            Meaningless on SpiNNaker,
            so we pretend we're the head node with a value of 0

            While there is a setter (for PyNN compatibility)
            any set call is ignored
        """
        return 0

    @mpi_rank.setter
    def mpi_rank(self, new_value: int) -> None:
        """
         sPyNNaker does not use this value meaningfully.

        :param new_value: Ignored
        """

    @property
    def num_processes(self) -> int:
        """
        The number of MPI worker processes.

        Settable.

        .. note::
            While there is a setter (for PyNN compatibility)
            any set call is ignored
        """
        return 1

    @num_processes.setter
    def num_processes(self, new_value: int) -> None:
        """
        sPyNNaker does not use this value meaningfully.

        :param new_value: Ignored
        """

    @property
    def dt(self) -> float:
        """
        The simulation time step in milliseconds.

       .. note::
            While there is a setter (for PyNN compatibility)
            any set call is ignored
        """
        return self.__writer.get_simulation_time_step_ms()

    @dt.setter
    def dt(self, _: Any) -> Never:
        """
        We do not support setting the time step except during setup.

        :raises NotImplementedError
        """
        raise NotImplementedError(
            "We do not support setting dt/ time step except during setup")

    @property
    def t(self) -> float:
        """
        The current simulation time in milliseconds.
        """
        return self.__writer.get_current_run_time_ms()

    @property
    def segment_counter(self) -> int:
        """
        The number of the current recording segment being generated.

        .. note::
            While there is a setter (for PyNN compatibility)
            any set call is ignored
        """
        return self.__writer.get_reset_number()

    @segment_counter.setter
    def segment_counter(self, _: Any) -> Never:
        """
        We do not support externally altering the segment counter

        raises: NotImplementedError
        """
        raise NotImplementedError(
            "We do not support externally altering the segment counter")

    @property
    def name(self) -> str:
        """
        The name of the simulator. Used to ensure PyNN recording neo
        blocks are correctly labelled.
        """
        return _version.NAME

    @property
    def recorders(self) -> Collection[Recorder]:
        """
        The recorders, used by the PyNN state object.
        """
        return self.__recorders

    @recorders.setter
    def recorders(self, new_value: Collection[Recorder]) -> None:
        """
        Setter for the internal recorders object

        :param new_value: the new value for the recorder
        """
        self.__recorders = set(new_value)

    def _set_up_timings(
            self, timestep: Optional[float], min_delay: Union[
                int, float, None],
            time_scale_factor: Optional[int]) -> None:
        """
        :param timestep: machine_time_Step in milliseconds
        :param min_delay:
        :param time_scale_factor:
        """

        # Get the standard values
        if timestep is None:
            self.__writer.set_up_timings_and_delay(
                timestep, time_scale_factor, min_delay)
        else:
            self.__writer.set_up_timings_and_delay(
                math.ceil(timestep * MICRO_TO_MILLISECOND_CONVERSION),
                time_scale_factor, min_delay)

        # Check the combination of machine time step and time scale factor
        if (self.__writer.get_simulation_time_step_ms() *
                self.__writer.get_time_scale_factor() < 0.1):
            logger.warning(
                "****************************************************")
            logger.warning(
                "*** The combination of simulation time step and  ***")
            logger.warning(
                "*** the machine time scale factor results in a   ***")
            logger.warning(
                "*** wall clock timer tick that is currently not  ***")
            logger.warning(
                "*** reliably supported by the SpiNNaker machine. ***")
            logger.warning(
                "****************************************************")

    def stop(self) -> None:
        """
        End running of the simulation. Notifying each Population of the end.
        """
        # pylint: disable=protected-access
        FecTimer.start_category(TimerCategory.SHUTTING_DOWN)
        for population in self.__writer.iterate_populations():
            population._end()

        super().stop()

    @staticmethod
    def register_binary_search_path(search_path: str) -> None:
        """
        Register an additional binary search path for executables.

        .. deprecated:: 7.0
            Use :py:meth:`SpynnakerDataView.register_binary_search_path`.

        :param search_path: absolute search path for binaries
        """
        SpynnakerDataView.register_binary_search_path(search_path)

    def _execute_write_neo_metadata(self) -> None:
        with FecTimer("Write Neo Metadata", TimerWork.OTHER):
            with NeoBufferDatabase() as db:
                db.write_segment_metadata()
                db.write_metadata()

    @overrides(AbstractSpinnakerBase._do_write_metadata)
    def _do_write_metadata(self) -> None:
        self._execute_write_neo_metadata()
        super()._do_write_metadata()

    def _execute_synapse_expander(self) -> None:
        with FecTimer("Synapse expander", TimerWork.SYNAPSE) as timer:
            if timer.skip_if_virtual_board():
                return
            synapse_expander()

    def _execute_neuron_expander(self) -> None:
        with FecTimer("Neuron expander", TimerWork.SYNAPSE) as timer:
            if timer.skip_if_virtual_board():
                return
            neuron_expander()

    def _execute_finish_connection_holders(self) -> None:
        with FecTimer("Finish connection holders", TimerWork.OTHER):
            finish_connection_holders()

    @overrides(AbstractSpinnakerBase._do_extra_load_algorithms)
    def _do_extra_load_algorithms(self) -> None:
        self._execute_neuron_expander()
        self._execute_synapse_expander()
        self._execute_finish_connection_holders()

    def _report_write_network_graph(self) -> None:
        with FecTimer("SpYNNakerNeuronGraphNetworkSpecificationReport",
                      TimerWork.REPORT) as timer:
            if timer.skip_if_cfg_false("Reports", "write_network_graph"):
                return
            spynnaker_neuron_graph_network_specification_report()

    @overrides(AbstractSpinnakerBase._do_extra_mapping_algorithms,
               extend_doc=False)
    def _do_extra_mapping_algorithms(self) -> None:
        self._report_write_network_graph()

    @overrides(AbstractSpinnakerBase._do_provenance_reports)
    def _do_provenance_reports(self) -> None:
        AbstractSpinnakerBase._do_provenance_reports(self)
        self._report_redundant_packet_count()

    def _report_redundant_packet_count(self) -> None:
        with FecTimer("Redundant packet count report",
                      TimerWork.REPORT) as timer:
            if timer.skip_if_cfg_false(
                    "Reports", "write_redundant_packet_count_report"):
                return
            redundant_packet_count_report()

    @overrides(AbstractSpinnakerBase._execute_splitter_selector)
    def _execute_splitter_selector(self) -> None:
        with FecTimer("Spynnaker splitter selector", TimerWork.OTHER):
            spynnaker_splitter_selector()

    @overrides(AbstractSpinnakerBase._execute_delay_support_adder,
               extend_doc=False)
    def _execute_delay_support_adder(self) -> None:
        """
        Runs, times and logs the DelaySupportAdder if required.
        """
        name = get_config_str_or_none("Mapping", "delay_support_adder")
        if name is None:
            return
        with FecTimer("DelaySupportAdder", TimerWork.OTHER):
            if name == "DelaySupportAdder":
                d_vertices, d_edges = delay_support_adder()
                for vertex in d_vertices:
                    self.__writer.add_vertex(vertex)
                for edge, partition_id in d_edges:
                    self.__writer.add_edge(edge, partition_id)
                return
            raise ConfigurationException(
                f"Unexpected cfg setting delay_support_adder: {name}")

    @overrides(AbstractSpinnakerBase._execute_buffer_extractor)
    def _execute_buffer_extractor(self) -> None:
        super()._execute_buffer_extractor()
        if not get_config_bool("Machine", "virtual_board"):
            with NeoBufferDatabase() as db:
                db.write_t_stop()



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from typing import Dict, List, Sequence, Tuple
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.application import (
    ApplicationEdge, ApplicationEdgePartition)
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import DelayExtensionException
from spynnaker.pyNN.extra_algorithms.splitter_components import (
    AbstractSpynnakerSplitterDelay, SplitterDelayVertexSlice)
from spynnaker.pyNN.models.neural_projections import (
    ProjectionApplicationEdge, DelayedApplicationEdge,
    DelayAfferentApplicationEdge)
from spynnaker.pyNN.models.utility_models.delays import DelayExtensionVertex
from spynnaker.pyNN.models.neuron import PopulationVertex


def delay_support_adder() -> Tuple[
        Sequence[DelayExtensionVertex], Sequence[Tuple[ApplicationEdge, str]]]:
    """
    Adds the delay extensions to the application graph, now that all the
    splitter objects have been set.

    :return: The delay vertices and delay edges that were added
    """
    adder = _DelaySupportAdder()
    return adder.add_delays()


class _DelaySupportAdder(object):
    """
    Adds delay extension vertices into the application graph as needed.
    """

    __slots__ = (
        "_app_to_delay_map",
        "_delay_post_edge_map",
        "_new_edges",
        "_new_vertices")

    def __init__(self) -> None:
        self._app_to_delay_map: Dict[
            ApplicationEdgePartition, DelayExtensionVertex] = dict()
        self._delay_post_edge_map: Dict[
            Tuple[DelayExtensionVertex, PopulationVertex, str],
            DelayedApplicationEdge] = dict()
        self._new_edges: List[Tuple[ApplicationEdge, str]] = list()
        self._new_vertices: List[DelayExtensionVertex] = list()

    def add_delays(self) -> Tuple[
            List[DelayExtensionVertex], List[Tuple[ApplicationEdge, str]]]:
        """
        Adds the delay extensions to the application graph, now that all the
        splitter objects have been set.

        :returns: Two independent lists, a list of the delay vertices and
           a list of delay edges and their partition id
        """
        progress = ProgressBar(1 + SpynnakerDataView.get_n_partitions(),
                               "Adding delay extensions as required")

        for vertex in SpynnakerDataView.get_vertices_by_type(
                DelayExtensionVertex):
            self._app_to_delay_map[vertex.partition] = vertex
            for edge in vertex.outgoing_edges:
                self._delay_post_edge_map[
                    vertex, edge.post_vertex,
                    vertex.partition.identifier] = edge
        progress.update(1)

        # go through all partitions.
        for app_outgoing_edge_partition in progress.over(
                SpynnakerDataView.iterate_partitions()):
            for app_edge in app_outgoing_edge_partition.edges:
                if isinstance(app_edge, ProjectionApplicationEdge):
                    self.__examine_edge_for_delays_to_add(
                        app_edge, app_outgoing_edge_partition)
        return self._new_vertices, self._new_edges

    def __examine_edge_for_delays_to_add(
            self, edge: ProjectionApplicationEdge,
            partition: ApplicationEdgePartition) -> None:
        """
        Look at a particular edge to see if it needs a delay vertex+edge
        inserted, and add it in if it does.
        """
        # figure the max delay and if we need a delay extension
        n_stages, steps_per_stage, need_delay_ext = self._check_delay_values(
            edge)

        # if we need a delay, add it to the app graph.
        if need_delay_ext:
            delay_app_vertex = self._create_delay_app_vertex_and_pre_edge(
                partition, edge, steps_per_stage, n_stages)

            # add the edge from the delay extension to the dest vertex
            self._create_post_delay_edge(
                delay_app_vertex, edge, partition.identifier)

    def _create_post_delay_edge(
            self, delay_app_vertex: DelayExtensionVertex,
            app_edge: ProjectionApplicationEdge,
            partition_id: str) -> None:
        """
        Creates the edge between delay extension and post vertex. Stores
        for future loading to the application graph when safe to do so.

        :param delay_app_vertex: delay extension vertex
        :param app_edge:
            the undelayed application edge this is associated with.
        :param partition_id: the partition id of the edge
        """
        # check for post edge
        delayed_edge = self._delay_post_edge_map.get(
            (delay_app_vertex, app_edge.post_vertex, partition_id), None)
        if delayed_edge is None:
            delay_edge = DelayedApplicationEdge(
                delay_app_vertex, app_edge.post_vertex,
                app_edge.synapse_information,
                label=(f"{app_edge.pre_vertex.label}_delayed_"
                       f"to_{app_edge.post_vertex.label}"),
                undelayed_edge=app_edge)
            self._delay_post_edge_map[
                delay_app_vertex, app_edge.post_vertex,
                partition_id] = delay_edge
            self._new_edges.append((delay_edge, partition_id))
            app_edge.delay_edge = delay_edge
            delay_app_vertex.add_outgoing_edge(delay_edge)

    def _create_delay_app_vertex_and_pre_edge(
            self, app_outgoing_edge_partition: ApplicationEdgePartition,
            app_edge: ProjectionApplicationEdge, delay_per_stage: int,
            n_delay_stages: int) -> DelayExtensionVertex:
        """
        Creates the delay extension application vertex and the edge from the
        source vertex to this delay extension. Adds to the graph, as safe to
        do so.

        :param app_outgoing_edge_partition:
            the original outgoing edge partition.
        :param app_edge: the undelayed application edge.
        :param delay_per_stage: delay for each delay stage
        :param n_delay_stages: the number of delay stages needed
        :return: the DelayExtensionAppVertex
        """
        # get delay extension vertex if it already exists.
        delay_app_vertex = self._app_to_delay_map.get(
            app_outgoing_edge_partition, None)
        if delay_app_vertex is None:
            # build delay app vertex
            delay_app_vertex = DelayExtensionVertex(
                app_outgoing_edge_partition, delay_per_stage, n_delay_stages,
                app_edge.pre_vertex.n_colour_bits,
                label=f"{app_edge.pre_vertex.label}_delayed")

            # set trackers
            delay_app_vertex.splitter = SplitterDelayVertexSlice()
            self._new_vertices.append(delay_app_vertex)
            self._app_to_delay_map[app_outgoing_edge_partition] = (
                delay_app_vertex)

            # build afferent app edge
            delay_pre_edge = DelayAfferentApplicationEdge(
                app_edge.pre_vertex, delay_app_vertex,
                label=f"{app_edge.pre_vertex.label}_to_DelayExtension")
            self._new_edges.append(
                (delay_pre_edge, app_outgoing_edge_partition.identifier))
        else:
            delay_app_vertex.set_new_n_delay_stages_and_delay_per_stage(
                n_delay_stages, delay_per_stage)
        return delay_app_vertex

    def _check_delay_values(self, app_edge: ProjectionApplicationEdge)\
            -> Tuple[int, int, bool]:
        """
        Checks the delay required from the user defined max, the max delay
        supported by the post vertex splitter and the delay Extensions.

        :param app_edge: the undelayed application edge
        :return: tuple(n_delay_stages, delay_steps_per_stage, extension_needed)
        """
        # get max delay required
        max_delay_needed_ms = max(
            synapse_info.synapse_dynamics.get_delay_maximum(
                synapse_info.connector, synapse_info)
            for synapse_info in app_edge.synapse_information)

        # get if the post vertex needs a delay extension
        post_splitter = app_edge.post_vertex.splitter
        if not isinstance(
                post_splitter, AbstractSpynnakerSplitterDelay):
            raise DelayExtensionException(
                f"The app vertex {app_edge.post_vertex} "
                f"with splitter {post_splitter} does not support delays "
                f"and yet requires a delay support for edge {app_edge}. "
                "Please use a Splitter which utilises the "
                "AbstractSpynnakerSplitterDelay interface.")
        max_delay_steps = app_edge.post_vertex.splitter.max_support_delay()
        time_step_ms = SpynnakerDataView.get_simulation_time_step_ms()
        max_delay_ms = max_delay_steps * time_step_ms

        # if does not need a delay extension, run away
        if max_delay_ms >= max_delay_needed_ms:
            return 0, max_delay_steps, False

        # Check post vertex is OK with getting a delay
        if not post_splitter.accepts_edges_from_delay_vertex():
            raise DelayExtensionException(
                f"The app vertex {app_edge.post_vertex} "
                f"with splitter {post_splitter} does not support delays "
                f"and yet requires a delay support for edge {app_edge}. "
                "Please use a Splitter which does not have "
                "accepts_edges_from_delay_vertex turned off.")

        # needs a delay extension, check can be supported with 1 delay
        # extension. We don't do more than 1 at the moment
        ext_provided_ms = DelayExtensionVertex.get_max_delay_ticks_supported(
                max_delay_steps) * time_step_ms
        total_delay_ms = ext_provided_ms + max_delay_ms
        if total_delay_ms < max_delay_needed_ms:
            raise DelayExtensionException(
                f"Edge:{app_edge.label} "
                f"has a max delay of {max_delay_needed_ms}. "
                f"But at a timestep of {time_step_ms} "
                f"the max delay supported is {total_delay_ms}")

        # return data for building delay extensions
        n_stages = int(math.ceil(max_delay_needed_ms / max_delay_ms)) - 1
        return n_stages, max_delay_steps, True



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyNN import common as pynn_common


class Assembly(pynn_common.Assembly):
    """
    A group of neurons, may be heterogeneous, in contrast to a Population
    where all the neurons are of the same type.

    :param populations: the populations or views to form the assembly out of
    :param kwargs: may contain `label` (a string describing the assembly)
    """



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Iterable, TYPE_CHECKING

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataSpecificationGenerator

from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamics)
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neuron import (
        PopulationMachineLocalOnlyCombinedVertex)


class AbstractLocalOnly(AbstractSynapseDynamics):
    """
    Processes synapses locally without the need for SDRAM.
    """

    def __init__(self, delay: WEIGHTS_DELAYS_IN):
        """
        :param delay:
            The delay used in the connection; by default 1 time step
        """
        # We don't have a weight here, it is in the connector
        super().__init__(delay=delay, weight=None)

    @abstractmethod
    def get_parameters_usage_in_bytes(
            self, n_atoms: int,
            incoming_projections: Iterable[Projection]) -> int:
        """
        :param n_atoms: The number of atoms in the vertex
        :param incoming_projections: The projections to get the size of
        :returns: The size of the parameters in bytes
        """
        raise NotImplementedError

    @abstractmethod
    def write_parameters(
            self, spec: DataSpecificationGenerator, region: int,
            machine_vertex: PopulationMachineLocalOnlyCombinedVertex,
            weight_scales: NDArray[floating]) -> None:
        """
        Write the parameters to the data specification for a vertex.

        :param spec: The specification to write to
        :param region: region ID to write to
        :param machine_vertex: The machine vertex being targeted
        :param weight_scales: Scale factors to apply to the weights
        """
        raise NotImplementedError

    @property
    def absolute_max_atoms_per_core(self) -> int:
        """
        The absolute maximum number of atoms per core.

        .. note::
            This is *not* constrained by the usual limits of the master
            population table.
        """
        # A bit of an estimate for these local-only systems, which don't use
        # the master population table and so don't have the same limit
        return 2048

    @property
    @overrides(AbstractSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return False



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Any, Dict, List, Optional, cast, TYPE_CHECKING, Union, Tuple

from spinn_utilities.overrides import overrides
from spynnaker.pyNN.models.neuron.implementations import NeuronImplStandard
from spynnaker.pyNN.models.neuron.abstract_pynn_neuron_model import (
    AbstractPyNNNeuronModel, _population_parameters as APNM_default_params)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.additional_inputs import (
        AbstractAdditionalInput)
    from spynnaker.pyNN.models.neuron.input_types import AbstractInputType
    from spynnaker.pyNN.models.neuron.neuron_models import NeuronModel
    from spynnaker.pyNN.models.neuron.synapse_types import AbstractSynapseType
    from spynnaker.pyNN.models.neuron.threshold_types import (
        AbstractThresholdType)
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        SplitterPopulationVertex)
    from .population_vertex import PopulationVertex

_population_parameters: Dict[str, Any] = dict(APNM_default_params)
_population_parameters["n_steps_per_timestep"] = 1


class AbstractPyNNNeuronModelStandard(AbstractPyNNNeuronModel):
    """
    A neuron model that follows the sPyNNaker standard composed model
    pattern for point neurons.
    """

    __slots__ = ()

    default_population_parameters = _population_parameters

    def __init__(
            self, model_name: str, binary: str, neuron_model: NeuronModel,
            input_type: AbstractInputType, synapse_type: AbstractSynapseType,
            threshold_type: AbstractThresholdType,
            additional_input_type: Optional[AbstractAdditionalInput] = None):
        """
        :param model_name: Name of the model.
        :param binary: Name of the implementation executable.
        :param neuron_model: The model of the neuron body
        :param input_type: The model of synaptic input types
        :param synapse_type: The model of the synapses' dynamics
        :param threshold_type: The model of the firing threshold
        :param additional_input_type:
            The model (if any) of additional environmental inputs
        """
        super().__init__(NeuronImplStandard(
            model_name, binary, neuron_model, input_type, synapse_type,
            threshold_type, additional_input_type))

    @overrides(AbstractPyNNNeuronModel.create_vertex)  # type: ignore[has-type]
    def create_vertex(
            self, n_neurons: int, label: str, *,
            spikes_per_second: Optional[float] = None,
            ring_buffer_sigma: Optional[float] = None,
            max_expected_summed_weight: Optional[List[float]] = None,
            incoming_spike_buffer_size: Optional[int] = None,
            drop_late_spikes: Optional[bool] = None,
            splitter: Optional[SplitterPopulationVertex] = None,
            seed: Optional[int] = None, n_colour_bits: Optional[int] = None,
            n_steps_per_timestep: int = 1,
            neurons_per_core: Optional[Union[int, Tuple[int, ...]]] = None,
            n_synapse_cores: Optional[int] = None,
            allow_delay_extensions: Optional[bool] = None) -> PopulationVertex:
        """
        :param n_steps_per_timestep:
        """
        cast(NeuronImplStandard,
             self._model).n_steps_per_timestep = n_steps_per_timestep
        return super().create_vertex(
            n_neurons=n_neurons, label=label,
            spikes_per_second=spikes_per_second,
            ring_buffer_sigma=ring_buffer_sigma,
            max_expected_summed_weight=max_expected_summed_weight,
            incoming_spike_buffer_size=incoming_spike_buffer_size,
            drop_late_spikes=drop_late_spikes,
            splitter=splitter, seed=seed, n_colour_bits=n_colour_bits,
            neurons_per_core=neurons_per_core, n_synapse_cores=n_synapse_cores,
            allow_delay_extensions=allow_delay_extensions)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Sequence, Union
from typing_extensions import TypeAlias
from pyNN.random import RandomDistribution
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.ranged_list import RangedList
from spinn_utilities.ranged.abstract_list import IdsType, T

# The type of things we consider to be a list of values
_ListType: TypeAlias = Union[Callable[[int], T], Sequence[T],
                             RandomDistribution]
# The type of value arguments in several places
_ValueType: TypeAlias = Optional[Union[T, _ListType]]


class SpynnakerRangedList(RangedList):
    """
    Adds support for :py:class:`~spynnaker.pyNN.RandomDistribution` to
    :py:class:`~spinn_utilities.ranged.RangedList`.
    """

    @overrides(RangedList.listness_check)
    def listness_check(self, value: _ValueType) -> bool:
        if isinstance(value, RandomDistribution):
            return True

        return super().listness_check(value)

    @overrides(RangedList.as_list)
    def as_list(
            self, value: _ListType, size: int,
            ids: Optional[IdsType] = None) -> List[T]:
        if isinstance(value, RandomDistribution):
            return value.next(n=size)

        return super().as_list(value, size, ids)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional
from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotRetinaResolution)


class AbstractPushBotRetinaDevice(
        AbstractSendMeMulticastCommandsVertex):
    """
    An abstraction of a silicon retina attached to a SpiNNaker system.
    """

    def __init__(self, protocol: MunichIoSpiNNakerLinkProtocol,
                 resolution: Optional[PushBotRetinaResolution]):
        """
        :param protocol:
        :param resolution:
        """
        self._protocol = protocol
        self._resolution = resolution

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # add mode command if not done already
        if not self._protocol.sent_mode_command():
            yield self._protocol.set_mode()

        # device specific commands
        yield self._protocol.disable_retina()

        retina_key = None
        if self._resolution is not None:
            retina_key = self._resolution.value
        yield self._protocol.set_retina_transmission(retina_key=retina_key)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        yield self._protocol.disable_retina()

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from neo import AnalogSignal, Block, Segment, SpikeTrain


def compare_spiketrain(spiketrain1: SpikeTrain, spiketrain2: SpikeTrain, *,
                       same_length: bool = True) -> None:
    """
    Checks two spike trains have the exact same data.

    :param spiketrain1: first spike train
    :param spiketrain2: second spike train
    :param same_length: Flag to indicate if the same length of data is
        held, i.e., all spikes up to the same time. If False allows one trains
        to have additional spikes after the first ends. This is used to
        compare data extracted part way with data extracted at the end.
    :raises AssertionError: If the spike trains are not equal
    """
    id1 = spiketrain1.annotations['source_index']
    id2 = spiketrain2.annotations['source_index']
    if id1 != id2:
        raise AssertionError(
            f"Different annotations['source_index'] found {id1} and {id2}")
    if same_length and len(spiketrain1) != len(spiketrain2):
        raise AssertionError(
            f"spiketrains1 has {len(spiketrain1)} spikes while spiketrains2 "
            f"has {len(spiketrain2)} for ID {id1}")
    for spike1, spike2 in zip(spiketrain1, spiketrain2):
        if spike1 != spike2:
            # print(id1, spiketrain1, spiketrain2)
            raise AssertionError(
                f"spike1 is {spike1} while spike2 is {spike2} for ID {id1}")


def compare_spiketrains(
        spiketrains1: List[SpikeTrain], spiketrains2: List[SpikeTrain], *,
        same_data: bool = True, same_length: bool = True) -> None:
    """
    Check two Lists of spike trains have the exact same data.

    :param spiketrains1:
        First list of spike trains to compare
    :param spiketrains2:
        Second list of spike trains to compare
    :param same_data: Flag to indicate if the same type of data is held,
        i.e., same spikes, v, gsyn_exc and gsyn_inh.
        If False allows one or both lists to be Empty.
        Even if False none empty lists must be the same length
    :param same_length: Flag to indicate if the same length of data is
        held, i.e., all spikes up to the same time. If False allows one trains
        to have additional spikes after the first ends. This is used to compare
        data extracted part way with data extracted at the end.
    :raises AssertionError: If the spike trains are not equal
    """
    if not same_data and (not spiketrains1 or not spiketrains2):
        return
    if len(spiketrains1) != len(spiketrains2):
        raise AssertionError(
            f"spiketrains1 has {len(spiketrains1)} spiketrains while "
            f"spiketrains2 fas {len(spiketrains2)} analogsignalarrays")
    for spiketrain1, spiketrain2 in zip(spiketrains1, spiketrains2):
        compare_spiketrain(spiketrain1, spiketrain2, same_length=same_length)


def compare_analogsignal(as1: AnalogSignal, as2: AnalogSignal, *,
                         same_length: bool = True) -> None:
    """
    Compares two analog signal objects to see if they are the same.

    :param as1:
        first analog signal holding list of individual analog signal objects
    :param as2:
        second analog signal holding list of individual analog signal objects
    :param same_length: Flag to indicate if the same length of data is
        held, i.e., all spikes up to the same time. If False allows one trains
        to have additional data after the first ends. This is used to compare
        data extracted part way with data extracted at the end.
    :raises AssertionError: If the analog signals are not equal
    """
    as1_index = as1.annotations["channel_names"]
    as2_index = as2.annotations["channel_names"]

    if as1.name != as2.name:
        raise AssertionError(
            f"analogsignalarray1 has name {as1.name} while "
            f"analogsignalarray1 has {as2.name}")

    if same_length and len(as1_index) != len(as2_index):
        raise AssertionError(
            f"channel_index 1 has len {len(as1_index)} while "
            f"channel_index 2 has {len(as2_index)} for {as1.name}")

    for id1, id2 in zip(as1_index, as2_index):
        if id1 != id2:
            raise AssertionError(
                f"ID 1 is {id1} while ID 2 is {id2} for {as1.name}")

    if same_length and len(as1.times) != len(as2.times):
        raise AssertionError(
            f"times 1 has len {len(as1.times)} while "
            f"times 2 has {len(as2.times)} for {as1.name}")

    for time1, time2 in zip(as1.times, as2.times):
        if time1 != time2:
            raise AssertionError(
                f"time 1 is {time1} while time 2 is {time2} for {as1.name}")

    if same_length and len(as1) != len(as2):
        raise AssertionError(
            f"analogsignal 1 has len {len(as1)} while "
            f"analogsignal 2 has {len(as2)} for {as1.name}")

    for signal1, signal2 in zip(as1, as2):
        # print(signal1, signal2)
        if len(signal1) != len(signal2):
            raise AssertionError(
                f"signal 1 has len {len(signal1)} while "
                f"signal 2 has {len(signal2)} for {as1.name}")
        for value1, value2 in zip(signal1, signal2):
            if value1 != value2:
                raise AssertionError(
                    f"value 1 is {value1} while "
                    f"value 2 is {value1} for {as1.name}")


def compare_segments(
        seg1: Segment, seg2: Segment, *, same_data: bool = True,
        same_length: bool = True) -> None:
    """
    COmpare two segments objects to see if they are the same.

    :param seg1: First Segment to check
    :param seg2: Second Segment to check
    :param same_data: Flag to indicate if the same type of data is held,
        i.e., same spikes, v, gsyn_exc and gsyn_inh.
        If False only data in both blocks is compared
    :param same_length: Flag to indicate if the same length of data is
        held, i.e., all spikes up to the same time. If False allows one trains
        to have additional data after the first ends. This is used to compare
        data extracted part way with data extracted at the end.
    :raises AssertionError: If the segments are not equal
    """
    compare_spiketrains(
        seg1.spiketrains, seg2.spiketrains,
        same_data=same_data, same_length=same_length)
    seg1_analogsignals = seg1.analogsignals
    seg2_analogsignals = seg2.analogsignals

    if same_data and len(seg1_analogsignals) != len(seg2_analogsignals):
        raise AssertionError(
            f"Segment1 has {len(seg1_analogsignals)} analogsignalarrays while "
            f"Segment2 has {len(seg1_analogsignals)} analogsignalarrays")

    for analogsignal1 in seg1_analogsignals:
        name = analogsignal1.name
        filtered = seg2.filter(name=name)
        if (not filtered) and same_data:
            raise AssertionError(
                f"Segment1 has {name} data while Segment2 does not")
        analogsignal2 = seg2.filter(name=name)[0]
        compare_analogsignal(analogsignal1, analogsignal2,
                             same_length=same_length)


def compare_blocks(
        neo1: Block, neo2: Block, *, same_runs: bool = True,
        same_data: bool = True, same_length: bool = True) -> None:
    """
    Compares two neo Blocks to see if they hold the same data.

    :param neo1: First block to check
    :param neo2: Second block to check
    :param same_runs: Flag to signal if blocks are the same length.
        If False extra segments in the larger block are ignored
    :param same_data: Flag to indicate if the same type of data is held,
        i.e., same spikes, v, gsyn_exc and gsyn_inh.
        If False only data in both blocks is compared
    :param same_length: Flag to indicate if the same length of data is
        held, i.e., all spikes up to the same time. If False allows one trains
        to have additional data after the first ends. This is used to compare
        data extracted part way with data extracted at the end.
    :raises AssertionError: If the blocks are not equal
    """
    if same_runs and len(neo1.segments) != len(neo2.segments):
        raise AssertionError(
            f"Block1 has {len(neo1.segments)} segments while "
            f"block2 has {len(neo2.segments)} segments")
    for seg1, seg2 in zip(neo1.segments, neo2.segments):
        compare_segments(seg1, seg2,
                         same_data=same_data, same_length=same_length)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
from scipy.stats import truncnorm
from spinn_utilities.overrides import overrides
from pyNN.random import RandomDistribution
from spynnaker.pyNN.utilities.random_stats import AbstractRandomStats


class RandomStatsNormalClippedImpl(AbstractRandomStats):
    """
    An implementation of AbstractRandomStats for normal distributions that
    are clipped to a boundary (redrawn).
    """

    def _get_params(self, dist: RandomDistribution) -> List[float]:
        low = ((dist.parameters['low'] - dist.parameters['mu']) /
               dist.parameters['sigma'])
        high = ((dist.parameters['high'] - dist.parameters['mu']) /
                dist.parameters['sigma'])
        return [low, high,
                dist.parameters['mu'], dist.parameters['sigma']]

    @overrides(AbstractRandomStats.cdf)
    def cdf(self, dist: RandomDistribution, v: float) -> float:
        return truncnorm.cdf(v, *self._get_params(dist))

    @overrides(AbstractRandomStats.ppf)
    def ppf(self, dist: RandomDistribution, p: float) -> float:
        return truncnorm.ppf(p, *self._get_params(dist))

    @overrides(AbstractRandomStats.mean)
    def mean(self, dist: RandomDistribution) -> float:
        return truncnorm.mean(*self._get_params(dist))

    @overrides(AbstractRandomStats.std)
    def std(self, dist: RandomDistribution) -> float:
        return truncnorm.std(*self._get_params(dist))

    @overrides(AbstractRandomStats.var)
    def var(self, dist: RandomDistribution) -> float:
        return truncnorm.var(*self._get_params(dist))

    @overrides(AbstractRandomStats.high)
    def high(self, distribution: RandomDistribution) -> Optional[float]:
        return distribution.parameters['high']

    @overrides(AbstractRandomStats.low)
    def low(self, distribution: RandomDistribution) -> Optional[float]:
        return distribution.parameters['low']



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.application import ApplicationVertex


@require_subclass(ApplicationVertex)
class AbstractHasDelayStages(object, metaclass=AbstractBase):
    """
    Indicates that this object (an application vertex) has delay stages that
    are used to increase the space required for bitfields in
    :py:func:`spynnaker.pyNN.utilities.bit_field_utilities.get_estimated_sdram_for_bit_field_region`
    """

    __slots__ = ()

    @property
    @abstractmethod
    def n_delay_stages(self) -> int:
        """
        The maximum number of delay stages required by any connection
        out of this delay extension vertex.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence import (
    TimingDependenceRecurrent as RecurrentRule,
    TimingDependenceSpikeNearestPair as SpikeNearestPairRule,
    TimingDependenceVogels2011 as Vogels2011Rule,
    TimingDependencePfisterSpikeTriplet as PfisterSpikeTriplet)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    SynapseDynamicsNeuromodulation as Neuromodulation,
    SynapseDynamicsWeightChangable as WeightChangeable,
    SynapseDynamicsWeightChanger as WeightChanger)
from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence import (
    WeightDependenceAdditiveTriplet)
from spynnaker.pyNN.models.neuron.builds import (
    IFCondExpStoc,
    IFCurrDelta as IFCurDelta,
    IFCurrExpCa2Adaptive,
    IFCurrDualExpBase as IF_curr_dual_exp,
    IzkCondExpBase as Izhikevich_cond,
    IzkCondDualExpBase as Izhikevich_cond_dual,
    IFCurrExpSEMDBase as IF_curr_exp_sEMD,
    IFCurrDeltaCa2Adaptive, StocExp, StocExpStable,  StocSigma, IFTruncDelta,
    IFCurrDeltaFixedProb)
from spynnaker.pyNN.models.neural_projections.connectors import (
    AllButMeConnector, OneToOneOffsetConnector)

# Variable rate poisson
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVariable

__all__ = [
    # sPyNNaker models not currently part of full pyNN
    'IFCurDelta', 'IFCurrExpCa2Adaptive', 'IFCondExpStoc',
    'Izhikevich_cond', 'IF_curr_dual_exp', 'IF_curr_exp_sEMD',
    'Izhikevich_cond_dual', 'IFCurrDeltaCa2Adaptive',

    # Neuromodulation synapse dynamics (Mantas Mikaitis)
    'Neuromodulation',

    # sPyNNaker 8 plastic stuff
    'WeightDependenceAdditiveTriplet',
    'PfisterSpikeTriplet',
    'SpikeNearestPairRule',
    'RecurrentRule', 'Vogels2011Rule',

    # Variable rate Poisson
    'SpikeSourcePoissonVariable',

    # Stochastic
    'StocExp', 'StocExpStable', 'StocSigma', 'IFCurrDeltaFixedProb',

    # Special
    'IFTruncDelta',

    # Connectors
    'AllButMeConnector', 'OneToOneOffsetConnector',

    # Weight changeable synapse dynamics
    'WeightChangeable', 'WeightChanger'
    ]
